@@ -1,13 +1,17 @@
 /* brushlib - The MyPaint Brush Library
- * Copyright (C) 2007-2008 Martin Renold <martinxyz@gmx.ch>
+ * Copyright (C) 2007-2011 Martin Renold <martinxyz@gmx.ch>
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
+ * Permission to use, copy, modify, and/or distribute this software for any
+ * purpose with or without fee is hereby granted, provided that the above
+ * copyright notice and this permission notice appear in all copies.
  *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY. See the COPYING file for more details.
+ * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+ * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+ * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+ * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+ * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
 #include <stdio.h>
@@ -61,6 +65,8 @@ private:
   // cached calculation results
   float speed_mapping_gamma[2], speed_mapping_m[2], speed_mapping_q[2];
 
+  bool reset_requested;
+
 public:
   Brush() {
     for (int i=0; i<BRUSH_SETTINGS_COUNT; i++) {
@@ -75,6 +81,8 @@ public:
     new_stroke();
 
     settings_base_values_have_changed();
+
+    reset_requested = true;
   }
 
   ~Brush() {
@@ -84,6 +92,11 @@ public:
     g_rand_free (rng); rng = NULL;
   }
 
+  void reset()
+  {
+    reset_requested = true;
+  }
+
   void new_stroke()
   {
     stroke_current_idling_time = 0;
@@ -91,22 +104,34 @@ public:
   }
 
   void set_base_value (int id, float value) {
-    g_assert (id >= 0 && id < BRUSH_SETTINGS_COUNT);
+    assert (id >= 0 && id < BRUSH_SETTINGS_COUNT);
     settings[id]->base_value = value;
 
     settings_base_values_have_changed ();
   }
 
   void set_mapping_n (int id, int input, int n) {
-    g_assert (id >= 0 && id < BRUSH_SETTINGS_COUNT);
+    assert (id >= 0 && id < BRUSH_SETTINGS_COUNT);
     settings[id]->set_n (input, n);
   }
 
   void set_mapping_point (int id, int input, int index, float x, float y) {
-    g_assert (id >= 0 && id < BRUSH_SETTINGS_COUNT);
+    assert (id >= 0 && id < BRUSH_SETTINGS_COUNT);
     settings[id]->set_point (input, index, x, y);
   }
 
+  float get_state (int i)
+  {
+    assert (i >= 0 && i < STATE_COUNT);
+    return states[i];
+  }
+
+  void set_state (int i, float value)
+  {
+    assert (i >= 0 && i < STATE_COUNT);
+    states[i] = value;
+  }
+
 private:
   // returns the fraction still left after t seconds
   float exp_decay (float T_const, float t)
@@ -166,7 +191,7 @@ private:
   // mappings in critical places or extremely few events per second.
   //
   // note: parameters are is dx/ddab, ..., dtime/ddab (dab is the number, 5.0 = 5th dab)
-  void update_states_and_setting_values (float step_dx, float step_dy, float step_dpressure, float step_dtime)
+  void update_states_and_setting_values (float step_dx, float step_dy, float step_dpressure, float step_declination, float step_ascension, float step_dtime)
   {
     float pressure;
     float inputs[INPUT_COUNT];
@@ -183,6 +208,9 @@ private:
     states[STATE_Y]        += step_dy;
     states[STATE_PRESSURE] += step_dpressure;
 
+    states[STATE_DECLINATION] += step_declination;
+    states[STATE_ASCENSION] += step_ascension;
+
     float base_radius = expf(settings[BRUSH_RADIUS_LOGARITHMIC]->base_value);
 
     // FIXME: does happen (interpolation problem?)
@@ -220,6 +248,8 @@ private:
     inputs[INPUT_RANDOM] = g_rand_double (rng);
     inputs[INPUT_STROKE] = MIN(states[STATE_STROKE], 1.0);
     inputs[INPUT_DIRECTION] = fmodf (atan2f (states[STATE_DIRECTION_DY], states[STATE_DIRECTION_DX])/(2*M_PI)*360 + 180.0, 180.0);
+    inputs[INPUT_TILT_DECLINATION] = states[STATE_DECLINATION];
+    inputs[INPUT_TILT_ASCENSION] = states[STATE_ASCENSION];
     inputs[INPUT_CUSTOM] = states[STATE_CUSTOM_INPUT];
     if (print_inputs) {
       g_print("press=% 4.3f, speed1=% 4.4f\tspeed2=% 4.4f\tstroke=% 4.3f\tcustom=% 4.3f\n", (double)inputs[INPUT_PRESSURE], (double)inputs[INPUT_SPEED1], (double)inputs[INPUT_SPEED2], (double)inputs[INPUT_STROKE], (double)inputs[INPUT_CUSTOM]);
@@ -383,8 +413,7 @@ private:
       radius_log  = settings_value[BRUSH_RADIUS_LOGARITHMIC];
       radius_log += rand_gauss (rng) * settings_value[BRUSH_RADIUS_BY_RANDOM];
       radius = expf(radius_log);
-      if (radius < ACTUAL_RADIUS_MIN) radius = ACTUAL_RADIUS_MIN;
-      if (radius > ACTUAL_RADIUS_MAX) radius = ACTUAL_RADIUS_MAX;
+      radius = CLAMP(radius, ACTUAL_RADIUS_MIN, ACTUAL_RADIUS_MAX);
       alpha_correction = states[STATE_ACTUAL_RADIUS] / radius;
       alpha_correction = SQR(alpha_correction);
       if (alpha_correction <= 1.0) {
@@ -425,13 +454,37 @@ private:
     if (settings_value[BRUSH_SMUDGE_LENGTH] < 1.0 and
         // optimization, since normal brushes have smudge_length == 0.5 without actually smudging
         (settings_value[BRUSH_SMUDGE] != 0.0 or not settings[BRUSH_SMUDGE]->is_constant())) {
+
       float fac = settings_value[BRUSH_SMUDGE_LENGTH];
-      if (fac < 0.0) fac = 0;
+      if (fac < 0.01) fac = 0.01;
       int px, py;
       px = ROUND(x);
       py = ROUND(y);
+
+      // Calling get_color() is almost as expensive as rendering a
+      // dab. Because of this we use the previous value if it is not
+      // expected to hurt quality too much. We call it at most every
+      // second dab.
       float r, g, b, a;
-      surface->get_color (px, py, radius, &r, &g, &b, &a);
+      states[STATE_LAST_GETCOLOR_RECENTNESS] *= fac;
+      if (states[STATE_LAST_GETCOLOR_RECENTNESS] < 0.5*fac) {
+        states[STATE_LAST_GETCOLOR_RECENTNESS] = 1.0;
+
+        float smudge_radius = radius * expf(settings_value[BRUSH_SMUDGE_RADIUS_LOG]);
+        smudge_radius = CLAMP(smudge_radius, ACTUAL_RADIUS_MIN, ACTUAL_RADIUS_MAX);
+        surface->get_color (px, py, smudge_radius, &r, &g, &b, &a);
+
+        states[STATE_LAST_GETCOLOR_R] = r;
+        states[STATE_LAST_GETCOLOR_G] = g;
+        states[STATE_LAST_GETCOLOR_B] = b;
+        states[STATE_LAST_GETCOLOR_A] = a;
+      } else {
+        r = states[STATE_LAST_GETCOLOR_R];
+        g = states[STATE_LAST_GETCOLOR_G];
+        b = states[STATE_LAST_GETCOLOR_B];
+        a = states[STATE_LAST_GETCOLOR_A];
+      }
+
       // updated the smudge color (stored with premultiplied alpha)
       states[STATE_SMUDGE_A ] = fac*states[STATE_SMUDGE_A ] + (1-fac)*a;
       // fix rounding errors
@@ -464,12 +517,35 @@ private:
       rgb_to_hsv_float (&color_h, &color_s, &color_v);
     }
 
-    float hardness = settings_value[BRUSH_HARDNESS];
+    float hardness = CLAMP(settings_value[BRUSH_HARDNESS], 0.0, 1.0);
+
+    // anti-aliasing attempt (works surprisingly well for ink brushes)
+    float current_fadeout_in_pixels = radius * (1.0 - hardness);
+    float min_fadeout_in_pixels = settings_value[BRUSH_ANTI_ALIASING];
+    if (current_fadeout_in_pixels < min_fadeout_in_pixels) {
+      // need to soften the brush (decrease hardness), but keep optical radius
+      // so we tune both radius and hardness, to get the desired fadeout_in_pixels
+      float current_optical_radius = radius - (1.0-hardness)*radius/2.0;
+
+      // Equation 1: (new fadeout must be equal to min_fadeout)
+      //   min_fadeout_in_pixels = radius_new*(1.0 - hardness_new)
+      // Equation 2: (optical radius must remain unchanged)
+      //   current_optical_radius = radius_new - (1.0-hardness_new)*radius_new/2.0
+      //
+      // Solved Equation 1 for hardness_new, using Equation 2: (thanks to mathomatic)
+      float hardness_new = ((current_optical_radius - (min_fadeout_in_pixels/2.0))/(current_optical_radius + (min_fadeout_in_pixels/2.0)));
+      // Using Equation 1:
+      float radius_new = (min_fadeout_in_pixels/(1.0 - hardness_new));
+
+      hardness = hardness_new;
+      radius = radius_new;
+    }
 
     // the functions below will CLAMP most inputs
     hsv_to_rgb_float (&color_h, &color_s, &color_v);
     return surface->draw_dab (x, y, radius, color_h, color_s, color_v, opaque, hardness, eraser_target_alpha,
-                              states[STATE_ACTUAL_ELLIPTICAL_DAB_RATIO], states[STATE_ACTUAL_ELLIPTICAL_DAB_ANGLE]);
+                              states[STATE_ACTUAL_ELLIPTICAL_DAB_RATIO], states[STATE_ACTUAL_ELLIPTICAL_DAB_ANGLE],
+                              settings_value[BRUSH_LOCK_ALPHA]);
   }
 
   // How many dabs will be drawn between the current and the next (x, y, pressure, +dt) position?
@@ -523,10 +599,37 @@ public:
   // - paints zero, one or several dabs
   // - decides whether the stroke is finished (for undo/redo)
   // returns true if the stroke is finished or empty
-  bool stroke_to (Surface * surface, float x, float y, float pressure, double dtime)
+  bool stroke_to (Surface * surface, float x, float y, float pressure, float xtilt, float ytilt, double dtime)
   {
     //printf("%f %f %f %f\n", (double)dtime, (double)x, (double)y, (double)pressure);
 
+    float tilt_ascension = 0.0;
+    float tilt_declination = 90.0;
+    if (xtilt != 0 || ytilt != 0) {
+      // shield us from insane tilt input
+      xtilt = CLAMP(xtilt, -1.0, 1.0);
+      ytilt = CLAMP(ytilt, -1.0, 1.0);
+      assert(isfinite(xtilt) && isfinite(ytilt));
+
+      tilt_ascension = 180.0*atan2(-xtilt, ytilt)/M_PI;
+      float e;
+      if (abs(xtilt) > abs(ytilt)) {
+        e = sqrt(1+ytilt*ytilt);
+      } else {
+        e = sqrt(1+xtilt*xtilt);
+      }
+      float rad = hypot(xtilt, ytilt);
+      float cos_alpha = rad/e;
+      if (cos_alpha >= 1.0) cos_alpha = 1.0; // fix numerical inaccuracy
+      tilt_declination = 180.0*acos(cos_alpha)/M_PI;
+
+      assert(isfinite(tilt_ascension));
+      assert(isfinite(tilt_declination));
+    }
+
+    // printf("xtilt %f, ytilt %f\n", (double)xtilt, (double)ytilt);
+    // printf("ascension %f, declination %f\n", (double)tilt_ascension, (double)tilt_declination);
+      
     pressure = CLAMP(pressure, 0.0, 1.0);
     if (!isfinite(x) || !isfinite(y) ||
         (x > 1e10 || y > 1e10 || x < -1e10 || y < -1e10)) {
@@ -545,7 +648,7 @@ public:
     if (dtime > 0.100 && pressure && states[STATE_PRESSURE] == 0) {
       // Workaround for tablets that don't report motion events without pressure.
       // This is to avoid linear interpolation of the pressure between two events.
-      stroke_to (surface, x, y, 0.0, dtime-0.0001);
+      stroke_to (surface, x, y, 0.0, 90.0, 0.0, dtime-0.0001);
       dtime = 0.0001;
     }
 
@@ -574,7 +677,8 @@ public:
     float dist_todo = count_dabs_to (x, y, pressure, dtime);
 
     //if (dtime > 5 || dist_todo > 300) {
-    if (dtime > 5) {
+    if (dtime > 5 || reset_requested) {
+      reset_requested = false;
 
       /*
         TODO:
@@ -612,6 +716,7 @@ public:
     double dtime_left = dtime;
 
     float step_dx, step_dy, step_dpressure, step_dtime;
+    float step_declination, step_ascension;
     while (dist_moved + dist_todo >= 1.0) { // there are dabs pending
       { // linear interpolation (nonlinear variant was too slow, see SVN log)
         float frac; // fraction of the remaining distance to move
@@ -627,10 +732,12 @@ public:
         step_dy        = frac * (y - states[STATE_Y]);
         step_dpressure = frac * (pressure - states[STATE_PRESSURE]);
         step_dtime     = frac * (dtime_left - 0.0);
+        step_declination = frac * (tilt_declination - states[STATE_DECLINATION]);
+        step_ascension   = frac * (tilt_ascension - states[STATE_ASCENSION]);
         // Though it looks different, time is interpolated exactly like x/y/pressure.
       }
     
-      update_states_and_setting_values (step_dx, step_dy, step_dpressure, step_dtime);
+      update_states_and_setting_values (step_dx, step_dy, step_dpressure, step_declination, step_ascension, step_dtime);
       bool painted_now = prepare_and_draw_dab (surface);
       if (painted_now) {
         painted = YES;
@@ -652,11 +759,13 @@ public:
       step_dx        = x - states[STATE_X];
       step_dy        = y - states[STATE_Y];
       step_dpressure = pressure - states[STATE_PRESSURE];
+      step_declination = tilt_declination - states[STATE_DECLINATION];
+      step_ascension = tilt_ascension - states[STATE_ASCENSION];
       step_dtime     = dtime_left;
     
       //dtime_left = 0; but that value is not used any more
 
-      update_states_and_setting_values (step_dx, step_dy, step_dpressure, step_dtime);
+      update_states_and_setting_values (step_dx, step_dy, step_dpressure, step_declination, step_ascension, step_dtime);
     }
 
     // save the fraction of a dab that is already done now
@@ -703,7 +812,7 @@ public:
         // Usually we have pressure==0 here. But some brushes can paint
         // nothing at full pressure (eg gappy lines, or a stroke that
         // fades out). In either case this is the prefered moment to split.
-        if (stroke_total_painting_time+stroke_current_idling_time > 1.2 + 5*pressure) {
+        if (stroke_total_painting_time+stroke_current_idling_time > 0.9 + 5*pressure) {
           return true;
         }
       }
@@ -711,38 +820,4 @@ public:
     return false;
   }
 
-  PyObject * get_state ()
-  {
-    npy_intp dims = {STATE_COUNT};
-    PyObject * data = PyArray_SimpleNew(1, &dims, NPY_FLOAT32);
-    npy_float32 * data_p = (npy_float32*)PyArray_DATA(data);
-    for (int i=0; i<STATE_COUNT; i++) {
-      data_p[i] = states[i];
-    }
-    return data;
-  }
-
-  void set_state (PyObject * data)
-  {
-    assert(PyArray_NDIM(data) == 1);
-    assert(PyArray_DIM(data, 0) == STATE_COUNT);
-    assert(PyArray_ISCARRAY(data));
-    npy_float32 * data_p = (npy_float32*)PyArray_DATA(data);
-    for (int i=0; i<STATE_COUNT; i++) {
-      states[i] = data_p[i];
-    }
-  }
-
-  double get_actual_radius() {
-
-    double base_radius = expf(settings[BRUSH_RADIUS_LOGARITHMIC]->base_value);
-    //double base_radius = states[STATE_ACTUAL_RADIUS];
-
-    double r = base_radius;
-
-    r += 2*base_radius*settings[BRUSH_OFFSET_BY_RANDOM]->base_value;
-    //r += 2*base_radius*settings_value[BRUSH_OFFSET_BY_RANDOM];
-
-    return r;
-  }
 };
