@@ -11,49 +11,9 @@ from os.path import join
 import gtk, gobject
 gdk = gtk.gdk
 from lib import brush, helpers, mypaintlib
-import filehandling, keyboard, brushmanager, windowing, document
-
-## TODO: move all window sizing stuff to windowing module
-##       while the keys below refer to fields in the Application, this will
-##       be ugly, however.
-
-# The window geometry that will be assumed at save time.
-SAVE_POS_GRAVITY = gdk.GRAVITY_NORTH_WEST
-
-# Window names and their default positions and visibilities. They interrelate,
-# so define them together.
-WINDOW_DEFAULTS = {
-    'drawWindow':           (None, True), # initial geom overridden, but set vis
-    # Colour choosers go in the top-right quadrant by default
-    'colorSamplerWindow':   ("220x220-50+100",  False),
-    'colorSelectionWindow': ("-250+150",        True), # positions strangely
-    # Brush-related dialogs go in the bottom right by default
-    'brushSelectionWindow': ("240x350-50-50",   True ),
-    'brushSettingsWindow':  ("400x400-325-50",  False),
-    # Preferences go "inside" the main window (TODO: just center-on-parent?)
-    'preferencesWindow':    ("+200+100",        False),
-    # Layer details in the bottom-left quadrant.
-    'layersWindow':         ("200x300+50-50",   False),
-    'backgroundWindow':     ("500x400-275+75", False),
-    # Debug menu
-    'inputTestWindow':      (None, False),
-}
-
-# The main drawWindow gets centre stage, sized around the default set of
-# visible utility windows. Sizing constraints for this:
-CENTRE_STAGE_CONSTRAINTS = (\
-    166, # left: leave some space for icons (but: OSX?)
-    50,  # top: avoid panel
-    75,  # bottom: avoid panel
-    10,  # right (small screens): ensure close button not covered at least
-    220, # right (big screens): don't overlap brushes or colour (much)
-    # "big screens" are anything 3 times wider than the big-right margin:
-    ## > Size. The Center Stage content should be at least twice as wide as
-    ## > whatever is in its side margins, and twice as tall as its top and
-    ## > bottom margins. (The user may change its size, but this is how it
-    ## > should be when the user first sees it.)
-    ## >     -- http://designinginterfaces.com/Center_Stage
-)
+import filehandling, keyboard, brushmanager, windowing, document, layout
+import colorhistory, brushmodifier
+import stock
 
 class Application: # singleton
     """
@@ -61,25 +21,53 @@ class Application: # singleton
     to be shared in the GUI. Its constructor is the last part of the
     initialization, called by main.py or by the testing scripts.
     """
-    def __init__(self, datapath, confpath, filenames):
+    def __init__(self, datapath, extradata, confpath, filenames):
+        """Construct, but do not run.
+
+        :`datapath`:
+            Usually ``$PREFIX/share/mypaint``. Where MyPaint should find its
+            app-specific read-only data, e.g. UI definition XML, backgrounds
+            and brush defintions.
+        :`extradata`:
+            Where to find the defaults for MyPaint's themeable UI icons. This
+            will be effectively used in addition to ``$XDG_DATA_DIRS`` for the
+            purposes of icon lookup. Normally it's ``$PREFIX/share``, to support
+            unusual installations outside the usual locations. It should contain
+            an ``icons/`` subdirectory.
+        :`confpath`:
+            Where the user's configuration is stored. ``$HOME/.mypaint`` is
+            typical on Unix-like OSes.
+        """
         self.confpath = confpath
         self.datapath = datapath
 
-        if not os.path.isdir(self.confpath):
-            os.mkdir(self.confpath)
-            print 'Created', self.confpath
+        # create config directory, and subdirs where the user might drop files
+        # TODO make scratchpad dir something pulled from preferences #PALETTE1
+        for d in ['', 'backgrounds', 'brushes', 'scratchpads']:
+            d = os.path.join(self.confpath, d)
+            if not os.path.isdir(d):
+                os.mkdir(d)
+                print 'Created', d
+
+        # Default location for our icons. The user's theme can override these.
+        icon_theme = gtk.icon_theme_get_default()
+        icon_theme.append_search_path(join(extradata, "icons"))
+
+        # Icon sanity check
+        if not icon_theme.has_icon('mypaint') \
+                or not icon_theme.has_icon('mypaint-tool-brush'):
+            print 'Error: Where have my icons gone?'
+            print 'Icon search path:', icon_theme.get_search_path()
+            print "Mypaint can't run sensibly without its icons; " \
+                + "please check your installation."
+            print 'see https://gna.org/bugs/?18460 for possible solutions'
+            sys.exit(1)
+        gtk.window_set_default_icon_name('mypaint')
+
+        stock.init_custom_stock_items()
 
         self.ui_manager = gtk.UIManager()
 
-        # if we are not installed, use the icons from the source
-        theme = gtk.icon_theme_get_default()
-        themedir_src = join(self.datapath, 'desktop/icons')
-        theme.prepend_search_path(themedir_src)
-        if not theme.has_icon('mypaint'):
-            print 'Warning: Where have all my icons gone?'
-            print 'Theme search path:', theme.get_search_path()
-        gtk.window_set_default_icon_name('mypaint')
-
         gdk.set_program_class('MyPaint')
 
         self.pixmaps = PixbufDirectory(join(self.datapath, 'pixmaps'))
@@ -87,42 +75,54 @@ class Application: # singleton
 
         # unmanaged main brush; always the same instance (we can attach settings_observers)
         # this brush is where temporary changes (color, size...) happen
-        self.brush = brush.Brush()
+        self.brush = brush.BrushInfo()
+        self.brush.load_defaults()
 
         self.preferences = {}
         self.load_settings()
-        self.brushmanager = brushmanager.BrushManager(join(datapath, 'brushes'), join(confpath, 'brushes'), self)
-        self.kbm = keyboard.KeyboardManager()
-        self.filehandler = filehandling.FileHandler(self)
-        self.doc = document.Document(self)
 
-        self.set_current_brush(self.brushmanager.selected_brush)
+        self.scratchpad_filename = ""
+        self.kbm = keyboard.KeyboardManager()
+        self.doc = document.Document(self)
+        self.scratchpad_doc = document.Document(self, leader=self.doc)
+        self.brushmanager = brushmanager.BrushManager(join(datapath, 'brushes'), join(confpath, 'brushes'), self)
+        self.filehandler = filehandling.FileHandler(self)
+        self.brushmodifier = brushmodifier.BrushModifier(self)
+
+
+        if not self.preferences.get("scratchpad.last_opened_scratchpad", None):
+            self.preferences["scratchpad.last_opened_scratchpad"] = self.filehandler.get_scratchpad_autosave()
+        self.scratchpad_filename = self.preferences["scratchpad.last_opened_scratchpad"]
+
         self.brush.set_color_hsv((0, 0, 0))
-        self.brushmanager.selected_brush_observers.append(self.brush_selected_cb)
         self.init_brush_adjustments()
 
-        self.user_subwindows = windowing.UserSubWindows(self)
-        self.window_names = ['drawWindow'] \
-          + [n for n in WINDOW_DEFAULTS.keys() if n != 'drawWindow']
-        for name in self.window_names:
-            module = __import__(name.lower(), globals(), locals(), [])
-            window = self.__dict__[name] = module.Window(self)
-            using_default = self.load_window_position(name, window)
-            if using_default:
-                if name == 'drawWindow':
-                    def on_map_event(win, ev, dw):
-                        windowing.centre_stage(win, *CENTRE_STAGE_CONSTRAINTS)
-                else:
-                    def on_map_event(win, ev, dw):
-                        windowing.move_to_monitor_of(win, dw)
-                window.connect('map-event', on_map_event, self.drawWindow)
+        self.ch = colorhistory.ColorHistory(self)
+
+        self.layout_manager = layout.LayoutManager(
+            prefs=self.preferences["layout.window_positions"],
+            factory=windowing.window_factory,
+            factory_opts=[self]  )
+        self.drawWindow = self.layout_manager.get_widget_by_role("main-window")
+        self.layout_manager.show_all()
 
         self.kbm.start_listening()
         self.filehandler.doc = self.doc
         self.filehandler.filename = None
         gtk.accel_map_load(join(self.confpath, 'accelmap.conf'))
 
-        def at_application_start(*trash):
+        # Load the background settings window.
+        # FIXME: this line shouldn't be needed, but we need to load this up
+        # front to get any non-default background that the user has configured
+        # from the preferences.
+        self.layout_manager.get_subwindow_by_role("backgroundWindow")
+
+        # And the brush settings window, or things like eraser mode will break.
+        # FIXME: brush_adjustments should not be dependent on this
+        self.layout_manager.get_subwindow_by_role("brushSettingsWindow")
+
+        def at_application_start(*junk):
+            self.brushmanager.select_initial_brush()
             if filenames:
                 # Open only the first file, no matter how many has been specified
                 # If the file does not exist just set it as the file to save to
@@ -132,6 +132,17 @@ class Application: # singleton
                 else:
                     self.filehandler.open_file(fn)
 
+            # Load last scratchpad
+            if not self.preferences["scratchpad.last_opened_scratchpad"]:
+                self.preferences["scratchpad.last_opened_scratchpad"] = self.filehandler.get_scratchpad_autosave()
+                self.scratchpad_filename = self.preferences["scratchpad.last_opened_scratchpad"]
+            if os.path.isfile(self.scratchpad_filename):
+                try:
+                    self.filehandler.open_scratchpad(self.scratchpad_filename)
+                except AttributeError, e:
+                    print "Scratchpad widget isn't initialised yet, so cannot centre"
+
+
             self.apply_settings()
             if not self.pressure_devices:
                 print 'No pressure sensitive devices found.'
@@ -147,16 +158,17 @@ class Application: # singleton
             f = open(settingspath, 'w')
             f.write(jsonstr)
             f.close()
+        self.brushmanager.save_brushes_for_devices()
+        self.brushmanager.save_brush_history()
+        self.filehandler.save_scratchpad(self.scratchpad_filename)
         save_config()
 
     def apply_settings(self):
         """Applies the current settings."""
         self.update_input_mapping()
         self.update_input_devices()
-        try:
-            self.preferencesWindow.update_ui()
-        except AttributeError:
-            pass
+        prefs_win = self.layout_manager.get_widget_by_role('preferencesWindow')
+        prefs_win.update_ui()
 
     def load_settings(self):
         '''Loads the settings from persistent storage. Uses defaults if
@@ -174,30 +186,131 @@ class Application: # singleton
         def get_json_config():
             settingspath = join(self.confpath, 'settings.json')
             jsonstr = open(settingspath).read()
-            return helpers.json_loads(jsonstr)
+            try:
+                return helpers.json_loads(jsonstr)
+            except Exception, e:
+                print "settings.json: %s" % (str(e),)
+                print "warning: failed to load settings.json, using defaults"
+                return {}
         if sys.platform == 'win32':
             import glib
             scrappre = join(glib.get_user_special_dir(glib.USER_DIRECTORY_DOCUMENTS).decode('utf-8'),'MyPaint','scrap')
-        else :
+        else:
             scrappre = '~/MyPaint/scrap'
         DEFAULT_CONFIG = {
             'saving.scrap_prefix': scrappre,
             'input.device_mode': 'screen',
             'input.global_pressure_mapping': [(0.0, 1.0), (1.0, 0.0)],
-            'input.enable_history_popup': True,
             'view.default_zoom': 1.0,
+            'view.high_quality_zoom': True,
+            'ui.hide_menubar_in_fullscreen': True,
+            'ui.hide_toolbar_in_fullscreen': True,
+            'ui.hide_subwindows_in_fullscreen': True,
+            'ui.parts': dict(main_toolbar=True, menubar=False),
+            'ui.toolbar_items': dict(
+                toolbar1_file=False,
+                toolbar1_edit=True,
+                toolbar1_view=False,
+                toolbar1_subwindows=True,
+            ),
             'saving.default_format': 'openraster',
             'brushmanager.selected_brush' : None,
             'brushmanager.selected_groups' : [],
+
+            "scratchpad.last_opened_scratchpad": "",
+
+            # Default window positions.
+            # See gui.layout.set_window_initial_position for the meanings
+            # of the common x, y, w, and h settings
+            "layout.window_positions": {
+
+                # Main window default size. Sidebar width is saved here
+                'main-window': dict(sbwidth=250, x=50, y=32, w=-50, h=-100),
+
+                # Tool windows. These can be undocked (floating=True) or set
+                # initially hidden (hidden=True), or be given an initial sidebar
+                # index (sbindex=<int>) or height in the sidebar (sbheight=<int>)
+                # Non-hidden entries determine the default set of tools.
+                'colorSelectionWindow': dict(
+                        sbindex=0, floating=True, hidden=True,
+                        x=-100, y=125, w=160, h=200, sbheight=200),
+                'colorSamplerWindow': dict(
+                        sbindex=1, floating=True, hidden=True,
+                        x=-270, y=125, w=200, h=275, sbheight=275),
+                'brushSelectionWindow': dict(
+                        sbindex=2, floating=True, hidden=True,
+                        x=-100, y=-150, w=250, h=350, sbheight=350),
+                'layersWindow': dict(
+                        sbindex=3, floating=True, hidden=True,
+                        x=-460, y=-150, w=200, h=200, sbheight=200),
+                'scratchWindow': dict(
+                        sbindex=4, floating=True, hidden=True,
+                        x=-555, y=125, w=300, h=250, sbheight=250),
+
+                # Non-tool subwindows. These cannot be docked, and are all
+                # intially hidden.
+                'brushSettingsWindow': dict(x=-460, y=-128, w=300, h=300),
+                'backgroundWindow': dict(),
+                'inputTestWindow': dict(),
+                'frameWindow': dict(),
+                'preferencesWindow': dict(),
+            },
         }
+        if sys.platform == 'win32':
+            # The Linux wacom driver inverts the button numbers of the
+            # pen flip button, because middle-click is the more useful
+            # action on Linux. However one of the two buttons is often
+            # accidentally hit with the thumb while painting. We want
+            # to assign panning to this button by default.
+            DEFAULT_CONFIG.update({
+                "input.button1_shift_action": 'straight_line',
+                "input.button1_ctrl_action":  'ColorPickerPopup',
+                "input.button3_action":       'pan_canvas',
+                "input.button3_shift_action": 'rotate_canvas',
+                "input.button3_ctrl_action":  'zoom_canvas',
+                "input.button2_action":       'ColorHistoryPopup',
+                "input.button2_shift_action": 'no_action',
+                "input.button2_ctrl_action":  'no_action',
+                })
+        else:
+            DEFAULT_CONFIG.update({
+                "input.button1_shift_action": 'straight_line',
+                "input.button1_ctrl_action":  'ColorPickerPopup',
+                "input.button2_action":       'pan_canvas',
+                "input.button2_shift_action": 'rotate_canvas',
+                "input.button2_ctrl_action":  'zoom_canvas',
+                "input.button3_action":       'ColorHistoryPopup',
+                "input.button3_shift_action": 'no_action',
+                "input.button3_ctrl_action":  'no_action',
+                })
+
+        window_pos = DEFAULT_CONFIG["layout.window_positions"]
+        self.window_names = window_pos.keys()
         self.preferences = DEFAULT_CONFIG
-        try: 
+        try:
             user_config = get_json_config()
         except IOError:
             user_config = get_legacy_config()
+        user_window_pos = user_config.get("layout.window_positions", {})
+        # note: .update() replaces the window position dict, but we want to update it
         self.preferences.update(user_config)
+        # update window_pos, and drop window names that don't exist any more
+        # (we need to drop them because otherwise we will try to show a non-existing window)
+        for role in self.window_names:
+            if role in user_window_pos:
+                window_pos[role] = user_window_pos[role]
+        self.preferences["layout.window_positions"] = window_pos
 
-    def init_brush_adjustments(self, ):
+    def add_action_group(self, ag):
+        self.ui_manager.insert_action_group(ag, -1)
+
+    def find_action(self, name):
+        for ag in self.ui_manager.get_action_groups():
+            result = ag.get_action(name)
+            if result is not None:
+                return result
+
+    def init_brush_adjustments(self):
         """Initializes all the brush adjustments for the current brush"""
         self.brush_adjustment = {}
         from brushlib import brushsettings
@@ -242,6 +355,11 @@ class Application: # singleton
                 # device.name is usually something like "wacom intuos3 6x8 pad" or just "pad"
                 print 'Ignoring "%s" (probably wacom keypad device)' % device.name
                 continue
+            if last_word == 'touchpad':
+                # eg. "SynPS/2 Synaptics TouchPad"
+                # Cannot paint at all, cannot select brushes, if we enable this one.
+                print 'Ignoring "%s" (probably laptop touchpad which screws up gtk+ if enabled)' % device.name
+                continue
             if last_word == 'cursor':
                 # this is a "normal" mouse and does not work in screen mode
                 print 'Ignoring "%s" (probably wacom mouse device)' % device.name
@@ -264,76 +382,9 @@ class Application: # singleton
                         device.set_mode(mode)
                     break
 
-    def set_current_brush(self, base_brush, settings_str=None):
-        if settings_str is None:
-            settings_str = base_brush.settings_str
-        self.brush.load_from_string(settings_str)
-
-    def brush_selected_cb(self, base_brush, settings_str=None):
-        assert base_brush is not self.brush
-        if base_brush is not None:
-            self.set_current_brush(base_brush, settings_str)
-
-    def hide_window_cb(self, window, event):
-        # used by some of the windows
-        window.hide()
-        return True
-
     def save_gui_config(self):
         gtk.accel_map_save(join(self.confpath, 'accelmap.conf'))
-        self.save_window_positions()
-	self.save_settings()
-
-    def save_window_positions(self):
-        f = open(join(self.confpath, 'windowpos.conf'), 'w')
-        f.write('# name visible x y width height\n')
-        for name in self.window_names:
-            window = self.__dict__[name]
-            x, y = window.get_position()
-            w, h = window.get_size()
-            gravity = window.get_gravity()
-            if gravity != SAVE_POS_GRAVITY:
-                # Then it was never mapped, and must still be using the
-                # defaults. Don't save position (which'd be wrong anyway).
-                continue
-            if hasattr(window, 'geometry_before_fullscreen'):
-                x, y, w, h = window.geometry_before_fullscreen
-            visible = window in self.user_subwindows.windows \
-                or window.get_property('visible') 
-            f.write('%s %s %d %d %d %d\n' % (name, visible, x, y, w, h))
-
-    def load_window_position(self, name, window):
-        geometry, visible = WINDOW_DEFAULTS.get(name, (None, False))
-        using_default = True
-        try:
-            for line in open(join(self.confpath, 'windowpos.conf')):
-                if line.startswith(name):
-                    parts = line.split()
-                    visible = parts[1] == 'True'
-                    x, y, w, h = [int(i) for i in parts[2:2+4]]
-                    geometry = '%dx%d+%d+%d' % (w, h, x, y)
-                    using_default = False
-                    break
-        except IOError:
-            pass
-        # Initial gravities can be all over the place. Fix aberrant ones up
-        # when the windows are safely on-screen so their position can be
-        # saved sanely. Doing this only when the window's mapped means that the
-        # window position stays where we specified in the defaults, and the
-        # window manager (should) compensate for us without the position
-        # changing.
-        if geometry is not None:
-            window.parse_geometry(geometry)
-            if using_default:
-                initial_gravity = window.get_gravity()
-                if initial_gravity != SAVE_POS_GRAVITY:
-                    def fix_gravity(w, event):
-                        if w.get_gravity() != SAVE_POS_GRAVITY:
-                            w.set_gravity(SAVE_POS_GRAVITY)
-                    window.connect("map-event", fix_gravity)
-        if visible:
-            window.show_all()
-        return using_default
+        self.save_settings()
 
     def message_dialog(self, text, type=gtk.MESSAGE_INFO, flags=0):
         """utility function to show a message/information dialog"""
@@ -350,7 +401,7 @@ class Application: # singleton
         root = screen.get_root_window()
         screen_w, screen_h = screen.get_width(), screen.get_height()
         display = widget.get_display()
-        screen_trash, x_root, y_root, modifiermask_trash = display.get_pointer()
+        screen_junk, x_root, y_root, modifiermask_trash = display.get_pointer()
         image = None
         x = x_root-size/2
         y = y_root-size/2
