@@ -16,22 +16,69 @@ from gtk import gdk # only for gdk.pixbuf
 from gettext import gettext as _
 import os, zipfile
 from os.path import basename
+import urllib
+import gobject
+from lib.brush import BrushInfo
+from warnings import warn
 
 preview_w = 128
 preview_h = 128
 
-DEFAULT_STARTUP_GROUP = 'Deevad'
+DEFAULT_STARTUP_GROUP = 'set#2'  # Suggestion only
+DEFAULT_BRUSH = 'deevad/artpen'  # TODO: phase out and use heuristics?
+DEFAULT_ERASER = 'deevad/stick'  # TODO: ---------------"--------------
 FOUND_BRUSHES_GROUP = 'lost&found'
 DELETED_BRUSH_GROUP = 'deleted'
 FAVORITES_BRUSH_GROUP = 'favorites'
+DEVBRUSH_NAME_PREFIX = "devbrush_"
+BRUSH_HISTORY_NAME_PREFIX = "history_"
+BRUSH_HISTORY_SIZE = 5
+NUM_BRUSHKEYS = 10
+
+def devbrush_quote(device_name, prefix=DEVBRUSH_NAME_PREFIX):
+    """
+    Quotes an arbitrary device name for use as the basename of a
+    device-specific brush.
+
+        >>> devbrush_quote(u'Heavy Metal Umlaut D\u00ebvice')
+        'devbrush_Heavy+Metal+Umlaut+D%C3%ABvice'
+        >>> devbrush_quote(u'unsafe/device\u005Cname') # U+005C == backslash
+        'devbrush_unsafe%2Fdevice%5Cname'
+
+    Hopefully this is OK for Windows, UNIX and Mac OS X names.
+    """
+    device_name = unicode(device_name)
+    u8bytes = device_name.encode("utf-8")
+    quoted = urllib.quote_plus(u8bytes, safe='')
+    return prefix + quoted
+
+def devbrush_unquote(devbrush_name, prefix=DEVBRUSH_NAME_PREFIX):
+    """
+    Unquotes the basename of a devbrush for use when matching device names.
+
+        >>> expected = "My sister was bitten by a m\u00f8\u00f8se..."
+        >>> quoted = 'devbrush_My+sister+was+bitten+by+a+m%5Cu00f8%5Cu00f8se...'
+        >>> devbrush_unquote(quoted) == expected
+        True
+    """
+    devbrush_name = str(devbrush_name)
+    assert devbrush_name.startswith(prefix)
+    quoted = devbrush_name[len(prefix):]
+    u8bytes = urllib.unquote_plus(quoted)
+    return unicode(u8bytes.decode("utf-8"))
 
 def translate_group_name(name):
-    d = {FOUND_BRUSHES_GROUP: _('lost&found'),
-         DELETED_BRUSH_GROUP: _('deleted'),
-         FAVORITES_BRUSH_GROUP: _('favorites'),
-         'ink': _('ink'),
-         'classic': _('classic'),
-         'experimental': _('experimental'),
+    d = {FOUND_BRUSHES_GROUP: _('Lost & Found'),
+         DELETED_BRUSH_GROUP: _('Deleted'),
+         FAVORITES_BRUSH_GROUP: _('Favorites'),
+         'ink': _('Ink'),
+         'classic': _('Classic'),
+         'set#1': _('Set#1'),
+         'set#2': _('Set#2'),
+         'set#3': _('Set#3'),
+         'set#4': _('Set#4'),
+         'set#5': _('Set#5'),
+         'experimental': _('Experimental'),
          }
     return d.get(name, name)
 
@@ -63,7 +110,7 @@ class BrushManager:
         self.user_brushpath = user_brushpath
         self.app = app
 
-        self.selected_brush = ManagedBrush(self)
+        self.selected_brush = None
         self.groups = {}
         self.contexts = []
         self.active_groups = []
@@ -73,33 +120,99 @@ class BrushManager:
 
         self.selected_brush_observers = []
         self.groups_observers = [] # for both self.groups and self.active_groups
-        self.brushes_observers = [] # for all bruslists inside groups
+        self.brushes_observers = [] # for all brushlists inside groups
 
         if not os.path.isdir(self.user_brushpath):
             os.mkdir(self.user_brushpath)
         self.load_groups()
 
-
+        # Retreive which groups were last open, or default to a nice/sane set.
         last_active_groups = self.app.preferences['brushmanager.selected_groups']
         if not last_active_groups:
-            last_active_groups = [DEFAULT_STARTUP_GROUP]
+            if DEFAULT_STARTUP_GROUP in self.groups:
+                last_active_groups = [DEFAULT_STARTUP_GROUP]
+            elif self.groups:
+                group_names = self.groups.keys()
+                group_names.sort()
+                last_active_groups = [group_names[0]]
+            else:
+                last_active_groups = []
         for group in reversed(last_active_groups):
             if group in self.groups:
                 brushes = self.get_group_brushes(group, make_active=True)
 
-        last_active_brush = self.app.preferences['brushmanager.selected_brush']
-        if last_active_brush is None:
-            brush = self.get_group_brushes(DEFAULT_STARTUP_GROUP)[0]
-        else:
-            brush = self.get_brush_by_name(last_active_brush)
-        self.select_brush(brush)
         self.brushes_observers.append(self.brushes_modified_cb)
 
+        self.app.doc.input_stroke_ended_observers.append(self.input_stroke_ended_cb)
+
+    def select_initial_brush(self):
+        initial_brush = None
+        # If we recorded which devbrush was last in use, restore it and assume
+        # that most of the time the user will continue to work with the same
+        # brush and its settings.
+        last_used_devbrush = self.app.preferences.get('devbrush.last_used', None)
+        initial_brush = self.brush_by_device.get(last_used_devbrush, None)
+        # Otherwise, initialise from the old selected_brush setting
+        if initial_brush is None:
+            last_active_name = self.app.preferences['brushmanager.selected_brush']
+            if last_active_name is not None:
+                initial_brush = self.get_brush_by_name(last_active_name)
+        # Fallback
+        if initial_brush is None:
+            initial_brush = self.get_default_brush()
+        self.select_brush(initial_brush)
+
+    def get_matching_brush(self, name=None, keywords=None,
+                           favored_group=DEFAULT_STARTUP_GROUP,
+                           fallback_eraser=0.0):
+        """Gets a brush robustly by name, by partial name, or a default.
+
+        If a brush named `name` exists, use that. Otherwise search though all
+        groups, `favored_group` first, for brushes with any of `keywords`
+        in their name. If that fails, construct a new default brush and use
+        a given value for its 'eraser' property.
+        """
+        if name is not None:
+            brush = self.get_brush_by_name(name)
+            if brush is not None:
+                return brush
+        if keywords is not None:
+            group_names = self.groups.keys()
+            group_names.sort()
+            if favored_group in self.groups:
+                group_names.remove(favored_group)
+                group_names.insert(0, favored_group)
+            for group_name in group_names:
+                for brush in self.groups[group_name]:
+                    for keyword in keywords:
+                        if keyword in brush.name:
+                            return brush
+        # Fallback
+        name = 'fallback-default'
+        if fallback_eraser != 0.0:
+            name += '-eraser'
+        brush = ManagedBrush(self, name)
+        brush.brushinfo.set_base_value("eraser", fallback_eraser)
+        return brush
+
+
+    def get_default_brush(self):
+        """Returns a suitable default drawing brush."""
+        return self.get_matching_brush(name=DEFAULT_BRUSH,
+                                keywords=["pencil", "charcoal", "sketch"])
+
+
+    def get_default_eraser(self):
+        """Returns a suitable default eraser brush."""
+        return self.get_matching_brush(name=DEFAULT_ERASER,
+                                keywords=["eraser", "kneaded", "smudge"],
+                                fallback_eraser=1.0)
+
+
+
     def load_groups(self):
-        for i in range(10):
-            c = ManagedBrush(self)
-            c.name = u'context%02d' % i
-            self.contexts.append(c)
+        self.contexts = [None for i in xrange(NUM_BRUSHKEYS)]
+        self.history = [None for i in xrange(BRUSH_HISTORY_SIZE)]
 
         brush_by_name = {}
         def get_brush(name):
@@ -183,15 +296,62 @@ class BrushManager:
                         l.append(name + '/' + name2)
             return l
 
+        # Distinguish between brushes in the brushlist and those that are not;
+        # handle lost-and-found ones.
         for name in listbrushes(self.stock_brushpath) + listbrushes(self.user_brushpath):
             b = get_brush(name)
+            b.in_brushlist = True
             if name.startswith('context'):
                 i = int(name[-2:])
                 self.contexts[i] = b
-                continue
-            if not [True for group in our.itervalues() if b in group]:
-                brushes = self.groups.setdefault(FOUND_BRUSHES_GROUP, [])
-                brushes.insert(0, b)
+                b.load_settings(retain_parent=True)
+                b.in_brushlist = False
+            elif name.startswith(DEVBRUSH_NAME_PREFIX):
+                device_name = devbrush_unquote(name)
+                self.brush_by_device[device_name] = b
+                b.load_settings(retain_parent=True)
+                b.in_brushlist = False
+            elif name.startswith(BRUSH_HISTORY_NAME_PREFIX):
+                i_str = name.replace(BRUSH_HISTORY_NAME_PREFIX, '')
+                i = int(i_str)
+                self.history[i] = b
+                b.load_settings(retain_parent=True)
+                b.in_brushlist = False
+            if b.in_brushlist:
+                if not [True for group in our.itervalues() if b in group]:
+                    brushes = self.groups.setdefault(FOUND_BRUSHES_GROUP, [])
+                    brushes.insert(0, b)
+
+        # Sensible defaults for brushkeys and history: clone the first few
+        # brushes from a normal group if we need to and if we can.
+        # Try the default startup group first.
+        default_group = self.groups.get(DEFAULT_STARTUP_GROUP, None)
+
+        # Otherwise, use the biggest group to minimise the chance
+        # of repetition.
+        if default_group is None:
+            groups_by_len = [(len(g),n,g) for n,g in self.groups.items()]
+            groups_by_len.sort()
+            _len, _name, default_group = groups_by_len[-1]
+
+        # Populate blank entries.
+        for i in xrange(NUM_BRUSHKEYS):
+            if self.contexts[i] is None:
+                idx = (i+9) % 10 # keyboard order
+                c_name = unicode('context%02d') % i
+                c = ManagedBrush(self, name=c_name, persistent=False)
+                group_idx = idx % len(default_group)
+                b = default_group[group_idx]
+                b.clone_into(c, c_name)
+                self.contexts[i] = c
+        for i in xrange(BRUSH_HISTORY_SIZE):
+            if self.history[i] is None:
+                h_name = unicode('%s%d') % (BRUSH_HISTORY_NAME_PREFIX, i)
+                h = ManagedBrush(self, name=h_name, persistent=False)
+                group_i = i % len(default_group)
+                b = default_group[group_i]
+                b.clone_into(h, h_name)
+                self.history[i] = h
 
         # clean up legacy stuff
         fn = os.path.join(self.user_brushpath, 'deleted.conf')
@@ -279,7 +439,6 @@ class BrushManager:
                 if brushname in new_brushes:
                     new_brushes.remove(brushname)
                     if b:
-                        b.load_preview()
                         existing_preview_pixbuf = b.preview
                         if do_ask:
                             answer = dialogs.confirm_rewrite_brush(window, brushname, existing_preview_pixbuf, preview_data)
@@ -318,6 +477,7 @@ class BrushManager:
                     preview_f.write(preview_data)
                     preview_f.close()
                     b.load()
+                    b.in_brushlist = True
                 # finally, add it to the group
                 if b not in managed_brushes:
                     managed_brushes.append(b)
@@ -342,6 +502,7 @@ class BrushManager:
 
     def get_brush_by_name(self, name):
         # slow method, should not be called too often
+        # FIXME: speed up, use a dict.
         for group, brushes in self.groups.iteritems():
             for b in brushes:
                 if b.name == name:
@@ -359,40 +520,116 @@ class BrushManager:
                 f.write(b.name.encode('utf-8') + '\n')
         f.close()
 
-    def select_brush(self, base_brush=None, settings_str=None):
+
+    def input_stroke_ended_cb(self, *junk):
+        """Update brush history at the end of an input stroke.
         """
-        Select a new ``ManagedBrush`` in the brush selector parts of the UI.
-        Also copy *either* ``brush``'s settings *or* the serialised settings
-        in ``settings_str`` into the brush settings parts of the UI and into
-        the ``Application`` instance's active brush.
+        b = self.app.brush
+        b_parent = b.get_string_property("parent_brush_name")
+        for i, h in enumerate(self.history):
+            h_parent = h.brushinfo.get_string_property("parent_brush_name")
+            # Possibly we should use a tighter equality check than this, but
+            # then we'd need icons showing modifications from the parent.
+            if b_parent == h_parent:
+                del self.history[i]
+                break
+        h = ManagedBrush(self, name=None, persistent=False)
+        h.brushinfo = b.clone()
+        h.preview = self.selected_brush.preview
+        self.history.append(h)
+        while len(self.history) > BRUSH_HISTORY_SIZE:
+            del self.history[0]
+        for i, h in enumerate(self.history):
+            h.name = u"%s%d" % (BRUSH_HISTORY_NAME_PREFIX, i)
 
-        Set both when you want to update settings while retaining a UI
-        reference to a base brush (basically keeping the parent highlighted but
-        updating the sliders for radius and so on).
 
-        Set just ``settings_str`` if that's all you have.
+    def select_brush(self, brush):
+        """Selects a ManagedBrush, highlights it, & updates the live brush."""
+        if brush is None:
+            brush = self.get_default_brush()
+        if brush.persistent and not brush.settings_loaded:   # XXX refactor
+            brush.load_settings()
 
-        See also ``BrushManager.selected_brush_observers``: most of the
-        legwork is delegated out to callbacks which follow the pattern
-        documented here.
-        """
-        if base_brush is None:
-            base_brush = ManagedBrush(self)
-        if base_brush.persistent and not base_brush.settings_loaded:
-            base_brush.load_settings()
-        assert isinstance(base_brush, ManagedBrush)
-        self.selected_brush = base_brush
-        self.app.preferences['brushmanager.selected_brush'] = base_brush.name
+        brushinfo = brush.brushinfo
+        if not brush.in_brushlist:
+            # select parent brush instead, but keep brushinfo
+            brush = self.get_parent_brush(brush=brush)
+            if not brush:
+                # no parent, select an empty brush instead
+                brush = ManagedBrush(self)
+
+        self.selected_brush = brush
+        self.app.preferences['brushmanager.selected_brush'] = brush.name
+        # Take care of updating the live brush, amongst other things
         for callback in self.selected_brush_observers:
-            callback(base_brush, settings_str)
+            callback(brush, brushinfo)
+
+
+    def get_parent_brush(self, brush=None, brushinfo=None):
+        """Gets the parent `ManagedBrush` for a brush or a `BrushInfo`.
+        """
+        if brush is not None:
+            if brush.persistent and not brush.settings_loaded:   # XXX refactor
+                brush.load_settings()
+            brushinfo = brush.brushinfo
+        if brushinfo is None:
+            raise RuntimeError, "One of `brush` or `brushinfo` must be defined."
+        parent_name = brushinfo.get_string_property("parent_brush_name")
+        if parent_name is None:
+            return None
+        else:
+            parent_brush = self.get_brush_by_name(parent_name)
+            if parent_brush is None:
+                return None
+            if parent_brush.persistent and not parent_brush.settings_loaded:  # XXX refactor
+                parent_brush.load_settings()
+            return parent_brush
+
+
+    def clone_selected_brush(self, name):
+        """
+        Creates a new ManagedBrush based on the selected
+        brush in the brushlist and the currently active lib.brush.
+        """
+        clone = ManagedBrush(self, name, persistent=False)
+        clone.brushinfo = self.app.brush.clone()
+        clone.preview = self.selected_brush.preview
+        parent = self.selected_brush.name
+        clone.brushinfo.set_string_property("parent_brush_name", parent)
+        return clone
+
+    def store_brush_for_device(self, device_name, managed_brush):
+        """
+        Records an existing ManagedBrush as associated with a given input device.
+
+        Normally the brush will be cloned first, since it will be given a new
+        name. However, if the ManagedBrush has a 'name' attribute of None, it
+        will *not* be cloned and just modified in place and stored.
+        """
+        brush = managed_brush
+        if brush.name is not None:
+            brush = brush.clone()
+        brush.name = unicode(devbrush_quote(device_name))
+        self.brush_by_device[device_name] = brush
+
+    def fetch_brush_for_device(self, device_name):
+        """
+        Fetches the brush associated with a particular input device name.
+        """
+        devbrush_name = devbrush_quote(device_name)
+        brush = self.brush_by_device.get(device_name, None)
+        return brush
+
+    def save_brushes_for_devices(self):
+        for device_name, devbrush in self.brush_by_device.iteritems():
+            devbrush.save()
+
+    def save_brush_history(self):
+        for brush in self.history:
+            brush.save()
 
     def set_active_groups(self, groups):
-        """Set active groups, loading them first if neccesary."""
-        for groupname in groups:
-            if not groupname in self.loaded_groups:
-                for brush in self.groups[groupname]:
-                    brush.load_preview()
-            self.loaded_groups.append(groupname)
+        """Set active groups."""
         self.active_groups = groups
         self.app.preferences['brushmanager.selected_groups'] = groups
         for f in self.groups_observers: f()
@@ -441,14 +678,12 @@ class ManagedBrush(object):
     '''Represents a brush, but cannot be selected or painted with directly.'''
     def __init__(self, brushmanager, name=None, persistent=False):
         self.bm = brushmanager
-        self.preview = None
+        self._preview = None
         self.name = name
-        self.settings_str = ''
-        self.persistent = persistent
-        """If True this brush is stored in the filesystem and 
-        not a context/picked brush."""
-        self.settings_loaded = False
-        """If True this brush is fully initialized, ready to paint with."""
+        self.brushinfo = BrushInfo()
+        self.persistent = persistent #: If True this brush is stored in the filesystem.
+        self.settings_loaded = False  #: If True this brush is fully initialized, ready to paint with.
+        self.in_brushlist = False  #: Set to True if this brush is known to be in the brushlist
 
         self.settings_mtime = None
         self.preview_mtime = None
@@ -457,6 +692,32 @@ class ManagedBrush(object):
             # we load the files later, but throw an exception now if they don't exist
             self.get_fileprefix()
 
+    # load preview pixbuf on demand
+    def get_preview(self):
+        if self._preview is None and self.name:
+            self._load_preview()
+        if self._preview is None:
+            self.preview = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, preview_w, preview_h)
+            self.preview.fill(0xffffffff) # white
+        return self._preview
+    def set_preview(self, pixbuf):
+        self._preview = pixbuf
+    preview = property(get_preview, set_preview)
+
+    def get_display_name(self):
+        """Gets a displayable name for the brush.
+        """
+        if self.in_brushlist:
+            dname = self.name
+        else:
+            if self.persistent and not self.settings_loaded:   # XXX refactor
+                self.load_settings()
+            dname = self.brushinfo.get_string_property("parent_brush_name")
+        if dname is None:
+            return _("Unknown Brush")
+        return dname.replace("_", " ")
+
+
     def get_fileprefix(self, saving=False):
         prefix = 'b'
         if os.path.realpath(self.bm.user_brushpath) == os.path.realpath(self.bm.stock_brushpath):
@@ -486,6 +747,22 @@ class ManagedBrush(object):
             raise IOError, 'brush "' + self.name + '" not found'
         return prefix
 
+    def clone(self, name):
+        "Creates a new brush with all the settings of this brush, assigning it a new name"
+        clone = ManagedBrush(self.bm)
+        self.clone_into(clone, name=name)
+        return clone
+
+    def clone_into(self, target, name):
+        "Copies all brush settings into another brush, giving it a new name"
+        if not self.settings_loaded:   # XXX refactor
+            self.load()
+        target.brushinfo = self.brushinfo.clone()
+        if self.in_brushlist:
+            target.brushinfo.set_string_property("parent_brush_name", self.name)
+        target.preview = self.preview
+        target.name = name
+
     def delete_from_disk(self):
         prefix = os.path.join(self.bm.user_brushpath, self.name)
         if os.path.isfile(prefix + '.myb'):
@@ -513,33 +790,41 @@ class ManagedBrush(object):
 
     def save(self):
         prefix = self.get_fileprefix(saving=True)
-        if self.preview is None:
-            self.preview = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, preview_w, preview_h)
-            self.preview.fill(0xffffffff) # white
         self.preview.save(prefix + '_prev.png', 'png')
-        open(prefix + '.myb', 'w').write(self.settings_str)
+        brushinfo = self.brushinfo.clone()
+        open(prefix + '.myb', 'w').write(brushinfo.save_to_string())
         self.remember_mtimes()
 
-    def load(self):
-        self.load_preview()
-        self.load_settings()
+    def load(self, retain_parent=False):
+        """Loads the brush's preview and settings from disk."""
+        if self.name is None:
+            warn("Attempt to load an unnamed brush, don't do that.",
+                 RuntimeWarning, 2)
+            return
+        self._load_preview()
+        self.load_settings(retain_parent)
 
-    def load_preview(self):
+    def _load_preview(self):
         """Loads the brush preview as pixbuf into the brush."""
+        assert self.name
         prefix = self.get_fileprefix()
 
         filename = prefix + '_prev.png'
         pixbuf = gdk.pixbuf_new_from_file(filename)
-        self.preview = pixbuf
+        self._preview = pixbuf
         self.remember_mtimes()
 
-    def load_settings(self):
+    def load_settings(self, retain_parent=False):
         """Loads the brush settings/dynamics from disk."""
         prefix = self.get_fileprefix()
         filename = prefix + '.myb'
-        self.settings_str = open(filename).read()
+        brushinfo_str = open(filename).read()
+        self.brushinfo.load_from_string(brushinfo_str)
         self.remember_mtimes()
         self.settings_loaded = True
+        if not retain_parent:
+            self.brushinfo.set_string_property("parent_brush_name", None)
+        self.persistent = True
 
     def reload_if_changed(self):
         if self.settings_mtime is None: return
@@ -550,3 +835,14 @@ class ManagedBrush(object):
         self.load()
         return True
 
+    def __str__(self):
+        if self.brushinfo.settings:
+            return "<ManagedBrush %s p=%s>" % (self.name, self.brushinfo.get_string_property("parent_brush_name"))
+        else:
+            return "<ManagedBrush %s (settings not loaded yet)>" % self.name
+
+
+if __name__ == '__main__':
+    import doctest
+    doctest.testmod()
+
