@@ -12,48 +12,55 @@ gdk = gtk.gdk
 from gettext import gettext as _
 
 import windowing
+import stock
 import pixbuflist, dialogs, brushmanager
+from elastic import ElasticExpander
 from brushlib import brushsettings
 
-class Window(windowing.SubWindow):
+class ToolWidget (gtk.VBox):
 
     EXPANDER_PREFS_KEY = "brushmanager.common_settings_expanded"
 
-    def __init__(self, app):
-        windowing.SubWindow.__init__(self, app)
-        self.last_selected_brush = None
+    stock_id = stock.TOOL_BRUSH
 
-        self.set_title(_('Brush selection'))
-        self.set_role('Brush selector')
-        self.connect('delete-event', self.app.hide_window_cb)
+    def __init__(self, app):
+        self.app = app
+        gtk.VBox.__init__(self)
+
+        self.last_selected_brush = None
 
         self.groupselector = GroupSelector(self.app)
         self.brushgroups = BrushGroupsList(self.app)
 
-        #main container
-        vbox = gtk.VBox()
-        self.add(vbox)
-
         self.scroll = scroll = gtk.ScrolledWindow()
         scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
         scroll.add_with_viewport(self.brushgroups)
 
-        is_expanded = bool(app.preferences.get(self.EXPANDER_PREFS_KEY, False))
-        expander = self.expander = gtk.Expander(label=None)
-        expander.set_expanded(is_expanded)
+        expander = self.expander = ElasticExpander(label=None)
         expander.add(get_common_settings_widget(app))
         expander.connect("notify::expanded", self.common_settings_expanded_cb)
+        self.expander_prefs_loaded = False
+        self.connect("show", self.show_cb)
 
-        al = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
-        al.add(self.groupselector)
-        al.set_padding(2,2,4,4)
-        vbox.pack_start(al, expand=False)
-        vbox.pack_start(gtk.HSeparator(), expand=False)
-        vbox.pack_start(scroll, expand=True)
-        vbox.pack_start(gtk.HSeparator(), expand=False)
-        vbox.pack_start(expander, expand=False, fill=False)
+        self.pack_start(self.groupselector, expand=False)
+        self.pack_start(gtk.HSeparator(), expand=False)
+        self.pack_start(scroll, expand=True)
+        self.pack_start(gtk.HSeparator(), expand=False)
+        self.pack_start(expander, expand=False, fill=False)
+
+        self.scroll.set_size_request(48, 48)
+        self.expander.set_size_request(48, -1)
+
+    def show_cb(self, widget):
+        assert not self.window
+        assert not self.expander_prefs_loaded
+        is_expanded = bool(self.app.preferences.get(self.EXPANDER_PREFS_KEY, False))
+        self.expander.set_expanded(is_expanded)
+        self.expander_prefs_loaded = True
 
     def common_settings_expanded_cb(self, expander, *junk):
+        if not self.expander_prefs_loaded:
+            return
         is_expanded = bool(expander.get_expanded())
         self.app.preferences[self.EXPANDER_PREFS_KEY] = is_expanded
 
@@ -64,14 +71,13 @@ def get_common_settings_widget(app):
     common_settings = [s for s in brushsettings.settings_visible if s.cname in cmn]
     settings_box = gtk.VBox()
 
-    def value_changed_cb(adj, index, app):
-        app.brush.settings[index].set_base_value(adj.get_value())
+    def value_changed_cb(adj, cname, app):
+        app.brush.set_base_value(cname, adj.get_value())
 
     def get_setting_widget(setting):
         """Return a widget to control a single setting"""
-
         adj = app.brush_adjustment[s.cname]
-        adj.connect('value-changed', value_changed_cb, s.index, app)
+        adj.connect('value-changed', value_changed_cb, s.cname, app)
 
         l = gtk.Label(s.name)
         l.set_alignment(0, 0.5)
@@ -100,6 +106,9 @@ class BrushList(pixbuflist.PixbufList):
         pixbuflist.PixbufList.__init__(self, self.brushes, 48, 48,
                                        namefunc = lambda x: x.name,
                                        pixbuffunc = lambda x: x.preview)
+        # Support device changing with the same event as that used
+        # for brush choice:
+        self.set_extension_events(gdk.EXTENSION_EVENTS_ALL)
         self.set_selected(self.bm.selected_brush)
         self.bm.brushes_observers.append(self.brushes_modified_cb)
         self.bm.selected_brush_observers.append(self.brush_selected_cb)
@@ -108,8 +117,8 @@ class BrushList(pixbuflist.PixbufList):
         if brushes is self.brushes:
             self.update()
 
-    def brush_selected_cb(self, base_brush, settings_str=None):
-        self.set_selected(base_brush)
+    def brush_selected_cb(self, managed_brush, brushinfo):
+        self.set_selected(managed_brush)
 
     def remove_brush(self, brush):
         self.brushes.remove(brush)
@@ -149,16 +158,12 @@ class BrushList(pixbuflist.PixbufList):
         return True
 
     def on_select(self, brush):
-        # keep the color setting
-        color = self.app.brush.get_color_hsv()
-
         # brush changed on harddisk?
         if brush.reload_if_changed():
             for brushes in self.bm.groups.itervalues():
                 for f in self.bm.brushes_observers: f(brushes)
 
         self.bm.select_brush(brush)
-        self.app.brush.set_color_hsv(color)
 
 class BrushGroupsList(gtk.VBox):
     def __init__(self, app):
@@ -192,6 +197,9 @@ class BrushGroupsList(gtk.VBox):
         self.show_all()
 
 class GroupSelector(gtk.DrawingArea):
+
+    VERTICAL_MARGIN = 2
+
     def __init__(self, app):
         gtk.DrawingArea.__init__(self)
 
@@ -222,34 +230,28 @@ class GroupSelector(gtk.DrawingArea):
         self.gtkstate_prelight_group = None
         self.gtkstate_active_group = None
         self.drag_target_group = None
-        self.set_tooltip_text(_('try right click, middle click or Ctrl click'))
+        self.set_tooltip_text(_('Try right click, middle click or Ctrl click'))
+        self.connect("size-request", self.on_size_request)
 
     def active_groups_changed_cb(self):
         self.queue_draw()
 
-    def expose_cb(self, widget, event):
-        cr = self.window.cairo_create()
-        width, height = self.window.get_size()
-
+    def lay_out_group_names(self, width):
+        "Typeset the group names into a new Pango layout at a given width"
+        self.ensure_style()
         style = self.get_style()
-
-        c = style.bg[gtk.STATE_NORMAL]
-        cr.set_source_rgb(c.red_float, c.green_float, c.blue_float)
-        cr.rectangle(0, 0, width, height)
-        cr.fill()
-
-        c = style.text[gtk.STATE_NORMAL]
-        cr.set_source_rgb(c.red_float, c.green_float, c.blue_float)
-        layout = cr.create_layout()
+        layout = pango.Layout(self.get_pango_context())
         layout.set_width(width*pango.SCALE)
-        layout.set_font_description(style.font_desc)
+        #layout.set_font_description(style.font_desc) # Needed?
 
         all_groups = list(sorted(self.bm.groups.keys()))
-
         idx = 0
         text = ''
         attr = pango.AttrList()
         self.idx2group = {}
+
+        # Pick separator chars. Platform-dependent, but assume Unicode
+        # for modern OSes first.
         pad_s = u"\u202f"  # NARROW NO-BREAK SPACE
         sp_s = pad_s + u"\u200b"  # ZERO WIDTH SPACE
 
@@ -267,7 +269,7 @@ class GroupSelector(gtk.DrawingArea):
             for c in s:
                 self.idx2group[idx] = group
                 idx += 1
-            
+
             # Note the difference in terminology here
             bg_state = fg_state = gtk.STATE_NORMAL
             if group == self.gtkstate_active_group: # activated the menu
@@ -285,7 +287,7 @@ class GroupSelector(gtk.DrawingArea):
 
             # always use the STATE_SELECTED fg if the group is visible
             if group in self.bm.active_groups:
-                fg_state = gtk.STATE_SELECTED 
+                fg_state = gtk.STATE_SELECTED
 
             c = style_bg[bg_state]
             attr.insert(pango.AttrBackground(c.red, c.green, c.blue, idx_start, idx))
@@ -297,22 +299,56 @@ class GroupSelector(gtk.DrawingArea):
 
         layout.set_text(text)
         layout.set_attributes(attr)
-        
+
+        leading = style.font_desc.get_size() / 6
+        layout.set_spacing(leading)
+        return layout
+
+    def on_size_request(self, widget, req):
+        parent = self.parent.parent
+        parent_width = parent.get_allocation().width
+        # The above is potentially the adopt "natural size" at first, but
+        # we should respect it if it's set. Might result in fewer redraws.
+        layout = self.lay_out_group_names(parent_width)
+        w, h = layout.get_pixel_size()
+        h += 2 * self.VERTICAL_MARGIN
+        req.width = -1
+        req.height = h
+
+    def expose_cb(self, widget, event):
+        cr = self.window.cairo_create()
+        alloc = self.get_allocation()
+        width = alloc.width
+        height = alloc.height
+
+        style = self.get_style()
+
+        c = style.bg[gtk.STATE_NORMAL]
+        cr.set_source_rgb(c.red_float, c.green_float, c.blue_float)
+        cr.rectangle(0, 0, width, height)
+        cr.fill()
+
+        c = style.text[gtk.STATE_NORMAL]
+        cr.set_source_rgb(c.red_float, c.green_float, c.blue_float)
+        layout = self.lay_out_group_names(width)
+
         leading = style.font_desc.get_size() / 6
         vmargin = leading // pango.SCALE
         layout.set_spacing(leading)
-        
-        w, h = layout.get_pixel_size()
-        h += 2*vmargin
-        cr.move_to(0, vmargin)
+
+        cr.move_to(0, self.VERTICAL_MARGIN)
         cr.show_layout(layout)
 
-        self.set_size_request(-1, h)
+        # Catch reflows, which maybe result in more or fewer rows.
+        self.set_size_request(-1, -1)  # "ask again, give me my natural size"
+        self.queue_resize_no_redraw()
 
         self.layout = layout
 
     def group_at(self, x, y):
         x, y = int(x), int(y) # avoid warning
+        if self.layout is None:
+            return None
         i, d = self.layout.xy_to_index(x*pango.SCALE, y*pango.SCALE)
         return self.idx2group.get(i)
 
@@ -368,6 +404,7 @@ class GroupSelector(gtk.DrawingArea):
                       (_("Delete group..."), self.delete_group_cb),
                       (_("Export group as brush package..."), self.export_group_cb),
                       ]
+        menu += [ (_("Download brushes (in web browser)"), self.app.drawWindow.download_brush_pack_cb) ]
         menu += [ (_("Import brush package..."), self.app.drawWindow.import_brush_pack_cb) ]
         for label, callback in menu:
             mi = gtk.MenuItem(label)
@@ -406,7 +443,7 @@ class GroupSelector(gtk.DrawingArea):
         if dialogs.confirm(self, _('Really delete group %s?') % brushmanager.translate_group_name(group)):
             self.bm.delete_group(group)
             if group in self.bm.groups:
-                dialogs.error(self, _('This group can not be deleted (try to make it empty first).'))
+                dialogs.error(self, _('This group can not be deleted (try to empty it first).'))
 
     def drag_data_received_cb(self, widget, context, x, y, selection, targetType, time):
         """
