@@ -1,12 +1,14 @@
 import gtk
 gdk = gtk.gdk
-from math import pi, sin, cos, sqrt, atan2
+from math import pi, sin, cos, sqrt, atan2, ceil
 import struct
 import cairo
 import windowing
+import stock
+from elastic import ElasticExpander
 from lib.helpers import rgb_to_hsv, hsv_to_rgb, clamp
-import colorhistory as ch
 from gettext import gettext as _
+import gobject
 
 CSTEP=0.007
 PADDING=4
@@ -38,6 +40,7 @@ class GColorSelector(gtk.DrawingArea):
         self.device_pressed = None
         self.grabbed = False
         self.set_size_request(110, 100)
+        self.has_hover_areas = False
 
     def test_move(self, x, y):
         """
@@ -60,6 +63,8 @@ class GColorSelector(gtk.DrawingArea):
         self.queue_draw()
 
     def motion(self,w, event):
+        if self.has_hover_areas:
+            self.update_hover_area(event.x, event.y)
         if not self.device_pressed:
             return
         if self.test_move(event.x, event.y):
@@ -81,13 +86,16 @@ class GColorSelector(gtk.DrawingArea):
     def on_select(self,color):
         pass
 
+    def update_tooltip(self, x, y):
+        """Updates the tooltip during motion, if tooltips are zoned."""
+        pass
+
     def get_color_at(self, x,y):
-        return self.color, False
+        return self.hsv, True
 
     def select_color_at(self, x,y):
-        try:
-            color, is_hsv = self.get_color_at(x,y)
-        except TypeError:
+        color, is_hsv = self.get_color_at(x,y)
+        if color is None:
             return
         if is_hsv:
             self.hsv = color
@@ -101,11 +109,11 @@ class GColorSelector(gtk.DrawingArea):
             self.app.doc.tdw.device_used(self.device_pressed)
             self.color = selected_color #... but *this* is what the user wants
         self.redraw_on_select()
-        self.on_select(self.color)
+        self.on_select(self.hsv)
 
-    def set_color(self, color):
-        self.color = color
-        self.hsv = rgb_to_hsv(*color)
+    def set_color(self, hsv):
+        self.hsv = hsv
+        self.color = hsv_to_rgb(*hsv)
         self.queue_draw()
 
     def get_color(self):
@@ -146,15 +154,18 @@ class RecentColors(gtk.HBox):
     def __init__(self, app):
         gtk.HBox.__init__(self)
         self.set_border_width(4)
-        self.N = N = ch.num_colors
+        self.N = N = app.ch.num_colors
         self.slots = []
-        for i,color in enumerate(reversed(ch.colors)):
-            slot = RectSlot(app, color=color)
+        for i in xrange(N):
+            slot = RectSlot(app)
             slot.on_select = self.slot_selected
             self.pack_start(slot, expand=True)
             self.slots.append(slot)
+        self.app = app
+        app.ch.color_pushed_observers.append(self.refill_slots)
+        self.set_tooltip_text(_("Recently used colors"))
+        self.refill_slots(None)
         self.show_all()
-        ch.on_color_pushed = self.refill_slots
 
     def slot_selected(self,color):
         self.on_select(color)
@@ -162,11 +173,11 @@ class RecentColors(gtk.HBox):
     def on_select(self,color):
         pass
 
-    def refill_slots(self, pushed_color):
-        for color,slot in zip(ch.colors, reversed(self.slots)):
-            slot.set_color(hsv_to_rgb(*color))
+    def refill_slots(self, *junk):
+        for hsv,slot in zip(self.app.ch.colors, reversed(self.slots)):
+            slot.set_color(hsv)
 
-    def set_color(self, color):
+    def set_color(self, hsv):
         pass
 
 CIRCLE_N = 12.0
@@ -178,7 +189,19 @@ AREA_SQUARE = 1  #: Central saturation/value square
 AREA_INSIDE = 2  #: Grey area inside the rings
 AREA_SAMPLE = 3  #: Inner ring of color sampler boxes (hue)
 AREA_CIRCLE = 4  #: Outer gradiated ring (hue)
-AREA_OUTSIDE = 5  #: Blank outer area, outer space
+AREA_COMPARE = 5  #: Current/previous comparison semicircles
+AREA_PICKER = 6  #: The picker icon's circle
+AREA_OUTSIDE = 7  #: Blank outer area, outer space
+
+CIRCLE_TOOLTIPS = \
+  { AREA_SQUARE: _('Change Saturation and Value'),
+    AREA_INSIDE: None,
+    AREA_SAMPLE: _('Harmony color switcher'),
+    AREA_CIRCLE: _('Change Hue'),
+    AREA_COMPARE: _('Current color vs. Last Used'),
+    AREA_PICKER: _('Pick a screen color'),
+    AREA_OUTSIDE: None,
+    }
 
 sq32 = sqrt(3)/2
 sq33 = sqrt(3)/3
@@ -190,6 +213,10 @@ def try_put(list, item):
         list.append(item)
 
 class CircleSelector(GColorSelector):
+
+    grab_mask = gdk.BUTTON_RELEASE_MASK | gdk.BUTTON1_MOTION_MASK
+    picking = False
+
     def __init__(self,app,color=(1,0,0)):
         GColorSelector.__init__(self,app)
         self.color = color
@@ -197,22 +224,120 @@ class CircleSelector(GColorSelector):
 
         self.samples = []      # [(h,s,v)] -- list of `harmonic' colors
         self.last_line = None  # 
+        app.ch.color_pushed_observers.append(self.color_pushed_cb)
 
-        # Whether to show harmonies
-        self.complimentary = False
-        self.triadic = False
-        self.double_comp = False
-        self.split_comp = False
-        self.analogous = False
-        self.square = False
+        self.has_hover_areas = True
+        self.hover_area = AREA_OUTSIDE
+        self.previous_hover_area = None
+        self.previous_hover_xy = None
+
+        self.picker_icons = []
+        self.connect("map-event", self.init_picker_icons)
+        self.connect("style-set", self.init_picker_icons)
+
+        self.picker_state = gtk.STATE_NORMAL
+        self.button_press = None
+
+    def on_button_press(self, widget, event):
+        area = self.area_at(event.x, event.y)
+        if area == AREA_PICKER:
+            self.button_press = event
+            self.picker_state = gtk.STATE_ACTIVE
+            self.queue_draw()
+            return True
+        return GColorSelector.on_button_press(self, widget, event)
+
+    def on_button_release(self, widget, event):
+        if self.picking:
+            self.app.pick_color_at_pointer(self)
+            self.picking = False
+            gdk.pointer_ungrab()
+            return True
+        elif self.button_press:
+            area = self.area_at(event.x, event.y)
+            if area == AREA_PICKER:
+                gobject.idle_add(self.begin_color_pick)
+            self.picker_state = gtk.STATE_NORMAL
+            self.queue_draw()
+            self.button_press = None
+            return True
+        return GColorSelector.on_button_release(self, widget, event)
+
+    def motion(self, widget, event):
+        if self.picking:
+            if event.state & gdk.BUTTON1_MASK:
+                self.app.pick_color_at_pointer(self)
+            return True
+        elif self.button_press:
+            area = self.area_at(event.x, event.y)
+            if area == AREA_PICKER:
+                if self.picker_state != gtk.STATE_ACTIVE:
+                    self.picker_state = gtk.STATE_ACTIVE
+                    self.queue_draw()
+            else:
+                if self.picker_state != gtk.STATE_NORMAL:
+                    self.picker_state = gtk.STATE_NORMAL
+                    self.queue_draw()
+            return True
+        return GColorSelector.motion(self, widget, event)
+
+    def begin_color_pick(self):
+        cursor = self.app.cursor_color_picker
+        result = gdk.pointer_grab(self.window, False, self.grab_mask, None, cursor)
+        if result == gdk.GRAB_SUCCESS:
+            self.picking = True
+
+    def init_picker_icons(self, *a, **kw):
+        sizes = [gtk.ICON_SIZE_SMALL_TOOLBAR,
+                 gtk.ICON_SIZE_LARGE_TOOLBAR,
+                 gtk.ICON_SIZE_BUTTON]
+                 # No bigger, to help avoid ugly scaling
+        pixbufs = []
+        for size in sizes:
+            pixbuf = self.render_icon(gtk.STOCK_COLOR_PICKER, size)
+            w = pixbuf.get_width()
+            h = pixbuf.get_height()
+            r = sqrt((w/2)**2 + (h/2)**2)
+            pixbufs.append((r, w, h, pixbuf, size))
+        pixbufs.sort() # in ascending order of size
+        self.picker_icons = pixbufs
+
+    def color_pushed_cb(self, pushed_color):
+        self.queue_draw()
 
     def has_harmonies_visible(self):
-        return self.complimentary or self.triadic or self.double_comp \
-          or self.split_comp or self.analogous or self.square
+        return self.app.preferences.get("colorsampler.complementary", False) \
+            or self.app.preferences.get("colorsampler.triadic", False) \
+            or self.app.preferences.get("colorsampler.double_comp", False) \
+            or self.app.preferences.get("colorsampler.split_comp", False) \
+            or self.app.preferences.get("colorsampler.analogous", False) \
+            or self.app.preferences.get("colorsampler.square", False)
 
     def get_previous_color(self):
         return self.color
 
+    def update_hover_area(self, x, y):
+        area = self.area_at(x, y)
+        old_area = area
+        if self.previous_hover_area is None:
+            old_area = None
+            tooltip = CIRCLE_TOOLTIPS.get(area, None)
+            self.set_tooltip_text(tooltip)
+            self.previous_hover_area = area, x, y
+        else:
+            old_area, old_x, old_y = self.previous_hover_area
+            if area != old_area or abs(old_x - x) > 20 or abs(old_y - y) > 20:
+                self.set_tooltip_text(None)
+                self.previous_hover_area = None
+        if area != old_area:
+            self.hover_area = area
+            if area == AREA_PICKER:
+                self.picker_state = gtk.STATE_PRELIGHT
+            else:
+                self.picker_state = gtk.STATE_NORMAL
+            if AREA_PICKER in (area, old_area):
+                self.queue_draw()
+
     def calc_line(self, angle):
         x1 = self.x0 + self.r2*cos(-angle)
         y1 = self.y0 + self.r2*sin(-angle)
@@ -236,25 +361,16 @@ class CircleSelector(GColorSelector):
         self.circle_img = None
         self.stroke_width = 0.01*M
 
-    def set_color(self, color, redraw=True):
-        self.color = color
-        h,s,v = rgb_to_hsv(*color)
-        old_h,old_s,old_v = self.hsv
-        self.hsv = h,s,v
+    def set_color(self, hsv, redraw=True):
+        self.hsv = hsv
+        self.color = hsv_to_rgb(*hsv)
         if redraw:
             self.queue_draw()
 
     def test_move(self, x, y):
         from_area = self.area_at(self.press_x,self.press_y)
         to_area = self.area_at(x, y)
-        if from_area == AREA_CIRCLE:
-            return to_area not in (AREA_SQUARE, AREA_INSIDE)
-        elif from_area == AREA_SQUARE:
-            return to_area not in (AREA_CIRCLE, AREA_OUTSIDE)
-        else:
-            return False
-        #return from_area == to_area
-        #    and from_area in [AREA_CIRCLE, AREA_SQUARE]
+        return from_area in (AREA_CIRCLE, AREA_SQUARE)
 
     def test_button_release(self, x, y):
         from_area = self.area_at(self.press_x,self.press_y)
@@ -280,11 +396,11 @@ class CircleSelector(GColorSelector):
             rx = self.x0 + (self.r2+3.0)*x1
             ry = self.y0 + (self.r2+3.0)*y1
         else:
-            tx = self.rd*x1
-            ty = self.rd*y1
-            m = self.m-1.0
-            rx = self.x0 + clamp(tx, -m, +m)
-            ry = self.y0 + clamp(ty, -m, +m)
+            m = self.m
+            dx = clamp(dx, -m, m)
+            dy = clamp(dy, -m, m)
+            rx = self.x0 + dx
+            ry = self.y0 + dy
         return rx,ry
 
     def move(self,x,y):
@@ -295,21 +411,35 @@ class CircleSelector(GColorSelector):
         dx = x-self.x0
         dy = y-self.y0
         d = sqrt(dx*dx+dy*dy)
-        if d > self.r3:
+        sq2 = sqrt(2.0)
+
+        # Two tools, the picker and the prev/current comparator
+        tool_M = self.M/2.0
+        tool_r = (sq2-1)*tool_M/(2*sq2)
+        tool_y0 = self.y0+tool_M-tool_r
+        cmp_x0 = self.x0+tool_M-tool_r
+        pick_x0 = self.x0-tool_M+tool_r
+        cmp_dx = x - cmp_x0
+        pick_dx = x - pick_x0
+        tool_dy = y - tool_y0
+
+        if tool_r > sqrt(cmp_dx*cmp_dx + tool_dy*tool_dy):
+            return AREA_COMPARE
+        if tool_r > sqrt(pick_dx*pick_dx + tool_dy*tool_dy):
+            return AREA_PICKER
+        elif d > self.r3:
             return AREA_OUTSIDE
-        elif self.r2 < d < self.r3:
+        elif self.r2 < d <= self.r3:
             return AREA_CIRCLE
-        elif self.rd < d < self.r2:
-            return AREA_SAMPLE
-        elif abs(dx)<self.m and abs(dy)<self.m:
+        elif abs(dx)<ceil(self.m) and abs(dy)<ceil(self.m):
             return AREA_SQUARE
+        elif self.rd < d <= self.r2:
+            return AREA_SAMPLE
         else:
             return AREA_INSIDE
 
     def get_color_at(self, x,y):
         area = self.area_at(x,y)
-        if not area:
-            return
         if area == AREA_CIRCLE:
             h,s,v = self.hsv
             h = 0.5 + 0.5*atan2(y-self.y0, self.x0-x)/pi
@@ -320,11 +450,13 @@ class CircleSelector(GColorSelector):
                 if a1-2*pi/CIRCLE_N < a < a1:
                     clr = self.simple_colors[i]
                     return clr, False
+            return None, False
         elif area == AREA_SQUARE:
             h,s,v = self.hsv
             s = (x-self.x0+self.m)/(2*self.m)
             v = (y-self.y0+self.m)/(2*self.m)
             return (h,s,1-v), True
+        return None, False
 
     def draw_square(self,width,height,radius):
         img = cairo.ImageSurface(cairo.FORMAT_ARGB32, width,height)
@@ -484,23 +616,23 @@ class CircleSelector(GColorSelector):
             cr.set_source_rgb(*NEUTRAL_DARK_GREY) # "lines" between samples
             cr.stroke()
             # Indicate harmonic colors
-            if self.triadic and i%(CIRCLE_N/3)==0:
+            if self.app.preferences.get("colorsampler.triadic", False) and i%(CIRCLE_N/3)==0:
                 self.small_triangle(cr, points_size, self.inv(clr), an, (self.r2+self.rd)/2)
                 try_put(self.samples, hsv)
-            if self.complimentary and i%(CIRCLE_N/2)==0:
+            if self.app.preferences.get("colorsampler.complementary", False) and i%(CIRCLE_N/2)==0:
                 self.small_circle(cr, points_size, self.inv(clr), an, (self.r2+self.rd)/2)
                 try_put(self.samples, hsv)
-            if self.square and i%(CIRCLE_N/4)==0:
+            if self.app.preferences.get("colorsampler.square", False) and i%(CIRCLE_N/4)==0:
                 self.small_square(cr, points_size, self.inv(clr), an, (self.r2+self.rd)/2)
                 try_put(self.samples, hsv)
 # FIXME: should this harmonies be expressed in terms of CIRCLE_N?
-            if self.double_comp and i in [0,2,6,8]:
+            if self.app.preferences.get("colorsampler.double_comp", False) and i in [0,2,6,8]:
                 self.small_rect_vert(cr, points_size, self.inv(clr), an, (self.r2+self.rd)/2)
                 try_put(self.samples, hsv)
-            if self.split_comp and i in [0,5,7]:
+            if self.app.preferences.get("colorsampler.split_comp", False) and i in [0,5,7]:
                 self.small_triangle_down(cr, points_size, self.inv(clr), an, (self.r2+self.rd)/2)
                 try_put(self.samples, hsv)
-            if self.analogous and i in [0,1,CIRCLE_N-1]:
+            if self.app.preferences.get("colorsampler.analogous", False) and i in [0,1,CIRCLE_N-1]:
                 self.small_rect(cr, points_size, self.inv(clr), an, (self.r2+self.rd)/2)
                 try_put(self.samples, hsv)
         # Fill the centre
@@ -534,7 +666,7 @@ class CircleSelector(GColorSelector):
         cr.stroke()
         cr.arc(self.x0,self.y0, self.r3, 0, 2*pi)
         cr.stroke()
-        
+
     def draw_circle(self, w, h):
         if self.circle_img:
             return self.circle_img
@@ -553,22 +685,62 @@ class CircleSelector(GColorSelector):
         self.circle_img = img
         return img
 
-    def draw_prev_vs_current_semicircles(self, cr):
+    def draw_comparison_semicircles(self, cr):
         sq2 = sqrt(2.0)
         M = self.M/2.0
         r = (sq2-1)*M/(2*sq2)
         x0 = self.x0+M-r
         y0 = self.y0+M-r
-        cr.set_source_rgb(*self.color)
-        cr.arc(x0, y0, 0.9*r, -pi/2, pi/2)
+        a0 = pi/4
+
+        prev_col = self.get_previous_color()
+        curr_col = self.color
+
+        cr.set_source_rgb(*NEUTRAL_MID_GREY)
+        cr.arc(x0, y0, 0.9*r, a0, a0+(2*pi))
         cr.fill()
-        cr.set_source_rgb(*self.get_previous_color())
-        cr.arc(x0, y0, 0.9*r, pi/2, 3*pi/2)
+
+        cr.set_source_rgb(*prev_col)
+        cr.arc(x0, y0, 0.8*r, a0, a0+(2*pi))
         cr.fill()
-        cr.arc(x0, y0, 0.9*r, 0, 2*pi)
-        cr.set_source_rgb(*NEUTRAL_DARK_GREY)
-        cr.set_line_width(0.75*self.stroke_width)
-        cr.stroke()
+
+        cr.set_source_rgb(*curr_col)
+        cr.arc(x0, y0, 0.8*r, a0+(pi/2), a0+(3*pi/2))
+        cr.fill()
+
+    def draw_picker_icon(self, cr):
+        sq2 = sqrt(2.0)
+        M = self.M/2.0
+        r = (sq2-1)*M/(2*sq2)
+        x0 = self.x0-M+r
+        y0 = self.y0+M-r
+        a0 = pi/4
+
+        # Pick the most suitable themed icon from our sorted cache.
+        # This should be the largest icon that fits in the tool circle,
+        # but we must use the smallest available if the circle is too small.
+        pixbuf = None
+        for pr, pw, ph, ppixbuf, psize in self.picker_icons:
+            if pixbuf is not None and pr > r:
+                break
+            pixbuf = ppixbuf
+            w = pw
+            h = ph
+
+        # Background
+        cr.save()
+        cr.arc(x0, y0, 0.9*r, a0, a0+(2*pi))
+        cr.clip()
+        bg = self.style.bg[self.picker_state]
+        bg_rgb = [bg.red_float, bg.green_float, bg.blue_float]
+        cr.set_source_rgb(*bg_rgb)
+        cr.paint()
+
+        # Picker icon
+        if pixbuf is not None:
+            cr.set_source_pixbuf(pixbuf, int(x0-w/2), int(y0-h/2))
+            cr.paint()
+        cr.restore()
 
     def draw(self,w,event):
         if not self.window:
@@ -582,7 +754,8 @@ class CircleSelector(GColorSelector):
         self.draw_circles(cr)
         cr.set_source_surface(self.draw_square(self.w, self.h, self.rd*0.95))
         cr.paint()
-        #self.draw_prev_vs_current_semicircles(cr)
+        self.draw_comparison_semicircles(cr)
+        self.draw_picker_icon(cr)
 
 class VSelector(GColorSelector):
     def __init__(self, app, color=(1,0,0), height=16):
@@ -593,7 +766,7 @@ class VSelector(GColorSelector):
 
     def get_color_at(self, x,y):
         h,s,v = self.hsv
-        v = x/self.w
+        v = clamp(x/self.w, 0.0, 1.0)
         return (h,s,v), True
 
     def move(self, x,y):
@@ -632,9 +805,9 @@ class VSelector(GColorSelector):
 class HSelector(VSelector):
     def get_color_at(self,x,y):
         h,s,v = self.hsv
-        h = x/self.w
+        h = clamp(x/self.w, 0.0, 1.0)
         return (h,s,v), True
-    
+
     def draw(self,w, event):
         if not self.window:
             return
@@ -656,7 +829,7 @@ class HSelector(VSelector):
 class SSelector(VSelector):
     def get_color_at(self, x,y):
         h,s,v = self.hsv
-        s = x/self.w
+        s = clamp(x/self.w, 0.0, 1.0)
         return (h,s,v), True
 
     def draw(self,w, event):
@@ -672,9 +845,9 @@ class SSelector(VSelector):
 class RSelector(VSelector):
     def get_color_at(self,x,y):
         r,g,b = self.color
-        r = x/self.w
+        r = clamp(x/self.w, 0.0, 1.0)
         return (r,g,b), False
-    
+
     def draw(self,w, event):
         if not self.window:
             return
@@ -687,9 +860,9 @@ class RSelector(VSelector):
 class GSelector(VSelector):
     def get_color_at(self,x,y):
         r,g,b = self.color
-        g = x/self.w
+        g = clamp(x/self.w, 0.0, 1.0)
         return (r,g,b), False
-    
+
     def draw(self,w, event):
         if not self.window:
             return
@@ -702,9 +875,9 @@ class GSelector(VSelector):
 class BSelector(VSelector):
     def get_color_at(self,x,y):
         r,g,b = self.color
-        b = x/self.w
+        b = clamp(x/self.w, 0.0, 1.0)
         return (r,g,b), False
-    
+
     def draw(self,w, event):
         if not self.window:
             return
@@ -748,27 +921,28 @@ class HSVSelector(gtk.VBox):
         self.vspin = vspin = make_spin(0,100, self.val_change)
         vbox.pack_start(vsel, expand=True)
         vbox.pack_start(vspin, expand=False)
-        
+
         self.pack_start(hbox, expand=False)
         self.pack_start(sbox, expand=False)
         self.pack_start(vbox, expand=False)
 
-    def user_selected_color(self, color):
-        self.set_color(color)
-        self.on_select(color)
+        self.set_tooltip_text(_("Change Hue, Saturation and Value"))
 
-    def set_color(self, color):
+    def user_selected_color(self, hsv):
+        self.set_color(hsv)
+        self.on_select(hsv)
+
+    def set_color(self, hsv):
         self.atomic = True
-        h,s,v = rgb_to_hsv(*color)
+        self.hsv = h,s,v = hsv
         self.hspin.set_value(h*359)
         self.sspin.set_value(s*100)
         self.vspin.set_value(v*100)
-        self.hsel.set_color(color)
-        self.ssel.set_color(color)
-        self.vsel.set_color(color)
+        self.hsel.set_color(hsv)
+        self.ssel.set_color(hsv)
+        self.vsel.set_color(hsv)
         self.atomic = False
-        self.color = color
-        self.hsv = rgb_to_hsv(*color)
+        self.color = hsv_to_rgb(*hsv)
 
     def on_select(self, color):
         pass
@@ -777,19 +951,19 @@ class HSVSelector(gtk.VBox):
         if self.atomic:
             return
         h,s,v = self.hsv
-        self.set_color(hsv_to_rgb(spin.get_value()/359., s,v))
+        self.set_color((spin.get_value()/359., s,v))
 
     def sat_change(self, spin):
         if self.atomic:
             return
         h,s,v = self.hsv
-        self.set_color(hsv_to_rgb(h, spin.get_value()/100., v))
+        self.set_color((h, spin.get_value()/100., v))
 
     def val_change(self, spin):
         if self.atomic:
             return
         h,s,v = self.hsv
-        self.set_color(hsv_to_rgb(h,s, spin.get_value()/100.))
+        self.set_color((h,s, spin.get_value()/100.))
 
 class RGBSelector(gtk.VBox):
     def __init__(self, app, color=(1.,0.,0)):
@@ -818,69 +992,67 @@ class RGBSelector(gtk.VBox):
         self.bspin = bspin = make_spin(0,255, self.b_change)
         bbox.pack_start(bsel, expand=True)
         bbox.pack_start(bspin, expand=False)
-        
+
         self.pack_start(rbox, expand=False)
         self.pack_start(gbox, expand=False)
         self.pack_start(bbox, expand=False)
 
-    def user_selected_color(self, color):
-        self.set_color(color)
-        self.on_select(color)
+        self.set_tooltip_text(_("Change Red, Green and Blue components"))
 
-    def set_color(self, color):
+    def user_selected_color(self, hsv):
+        self.set_color(hsv)
+        self.on_select(hsv)
+
+    def set_color(self, hsv):
         self.atomic = True
-        r,g,b = color
+        self.color = r,g,b = hsv_to_rgb(*hsv)
         self.rspin.set_value(r*255)
         self.gspin.set_value(g*255)
         self.bspin.set_value(b*255)
-        self.rsel.set_color(color)
-        self.gsel.set_color(color)
-        self.bsel.set_color(color)
+        self.rsel.set_color(hsv)
+        self.gsel.set_color(hsv)
+        self.bsel.set_color(hsv)
         self.atomic = False
-        self.color = color
-        self.hsv = rgb_to_hsv(*color)
+        self.hsv = hsv
 
-    def on_select(self, color):
+    def on_select(self, hsv):
         pass
 
     def r_change(self, spin):
         if self.atomic:
             return
         r,g,b = self.color
-        self.set_color((spin.get_value()/255., g,b))
+        self.set_color(rgb_to_hsv(spin.get_value()/255., g,b))
 
     def g_change(self, spin):
         if self.atomic:
             return
         r,g,b = self.color
-        self.set_color((r, spin.get_value()/255., b))
+        self.set_color(rgb_to_hsv(r, spin.get_value()/255., b))
 
     def b_change(self, spin):
         if self.atomic:
             return
         r,g,b = self.color
-        self.set_color((r,g, spin.get_value()/255.))
+        self.set_color(rgb_to_hsv(r,g, spin.get_value()/255.))
 
 class Selector(gtk.VBox):
 
-    CIRCLE_MIN_SIZE = (200, 200)
+    CIRCLE_MIN_SIZE = (150, 150)
 
-    def __init__(self, app, window):
+    def __init__(self, app):
         gtk.VBox.__init__(self)
         self.app = app
-        hbox = gtk.HBox()
-        self.pack_start(hbox, expand=True)
-        vbox = gtk.VBox()
-        hbox.pack_start(vbox,expand=True)
         self.recent = RecentColors(app)
         self.circle = CircleSelector(app)
         self.circle.set_size_request(*self.CIRCLE_MIN_SIZE)
-        vbox.pack_start(self.circle, expand=True)
 
-        self.window_ = window
-        self.window_.connect("configure-event", self.window_configured_cb)
-        self.window_.connect("map-event", self.window_mapped_cb)
-        self.window_mapped = False
+        # The colour circle seems to need its own window to draw sanely when
+        # packed into a Tool wrapper, which probably indicates that something's
+        # amiss somewhere.
+        evbox = gtk.EventBox()
+        evbox.add(self.circle)
+        self.pack_start(evbox, expand=True)
 
         self.rgb_selector = RGBSelector(app)
         self.hsv_selector = HSVSelector(app)
@@ -891,36 +1063,38 @@ class Selector(gtk.VBox):
         nb.append_page(self.hsv_selector, gtk.Label(_('HSV')))
 
         # Colour history
-        self.exp_history = expander = gtk.Expander(_('Colors history'))
+        self.exp_history = expander = ElasticExpander(_('Color history'))
         expander.set_spacing(6)
         expander.add(self.recent)
         expander.connect("notify::expanded", self.expander_expanded_cb, 'history')
         self.pack_start(expander, expand=False)
 
         # Colour details
-        self.exp_details = expander = gtk.Expander(_('Details'))
+        self.exp_details = expander = ElasticExpander(_('Details'))
         expander.set_spacing(6)
         expander.add(nb)
         expander.connect("notify::expanded", self.expander_expanded_cb, 'details')
         self.pack_start(expander, expand=False)
 
         # Colour scheme harmonies
-        def harmony_checkbox(attr, label):
+        def harmony_checkbox(attr, label, tooltip=None):
             cb = gtk.CheckButton(label)
+            pref = "colorsampler.%s" % (attr,)
+            cb.set_active(self.app.preferences.get(pref, False))
             cb.connect('toggled', self.harmony_toggled, attr)
+            if tooltip is not None:
+                cb.set_tooltip_text(tooltip)
             vbox2.pack_start(cb, expand=False)
 
-        self.exp_config = expander = gtk.Expander(_('Harmonies'))
+        self.exp_config = expander = ElasticExpander(_('Harmonies'))
         vbox2 = gtk.VBox()
-        harmony_checkbox('analogous', _('Analogous'))
-        harmony_checkbox('complimentary', _('Complimentary color'))
-        harmony_checkbox('split_comp', _('Split complimentary'))
-        harmony_checkbox('double_comp', _('Double complimentary'))
-        harmony_checkbox('square', _('Square'))
-        harmony_checkbox('triadic', _('Triadic'))
+        harmony_checkbox('analogous', _('Analogous'), _("Three nearby hues on the color wheel.\nOften found in nature, frequently pleasing to the eye."))
+        harmony_checkbox('complementary', _('Complementary'), _("Two opposite hues on the color wheel.\nVibrant, and maximally contrasting."))
+        harmony_checkbox('split_comp', _('Split complementary'), _("Two hues next to the current hue's complement.\nContrasting, but adds a possibly pleasing harmony."))
+        harmony_checkbox('double_comp', _('Double complementary'), _("Four hues in two complementary pairs."))
+        harmony_checkbox('square', _('Square'), _("Four equally-spaced hues"))
+        harmony_checkbox('triadic', _('Triadic'), _("Three equally-spaced hues.\nVibrant with equal tension."))
 
-        frame1 = gtk.Frame(_('Select harmonies'))
-        frame1.add(vbox2)
         vbox3 = gtk.VBox()
         cb_sv = gtk.CheckButton(_('Change value/saturation'))
         cb_sv.set_active(True)
@@ -932,9 +1106,7 @@ class Selector(gtk.VBox):
         vbox3.pack_start(cb_sv, expand=False)
         vbox3.pack_start(cb_opposite, expand=False)
         vbox3.pack_start(cb_neg, expand=False)
-        vbox_exp = gtk.VBox()
-        vbox_exp.pack_start(frame1)
-        expander.add(vbox_exp)
+        expander.add(vbox2)
         expander.connect("notify::expanded", self.expander_expanded_cb, 'harmonies')
         self.pack_start(expander, expand=False)
 
@@ -947,89 +1119,101 @@ class Selector(gtk.VBox):
         self.opposite_blends = False
         self.negative_blends = False
 
-        self.connect('drag_data_received',self.drag_data)
-        self.drag_dest_set(gtk.DEST_DEFAULT_MOTION | gtk.DEST_DEFAULT_HIGHLIGHT | gtk.DEST_DEFAULT_DROP,
-                 [("application/x-color",0,80)],
-                 gtk.gdk.ACTION_COPY)
+        self.expander_prefs_loaded = False
+        self.connect("show", self.show_cb)
 
     def toggle_blend(self, checkbox, name):
         attr = name+'_blends'
         setattr(self, attr, not getattr(self, attr))
 
     def harmony_toggled(self, checkbox, attr):
-        setattr(self.circle, attr, not getattr(self.circle, attr))
+        pref = "colorsampler.%s" % (attr,)
+        self.app.preferences[pref] = checkbox.get_active()
         self.circle.configure_calc()
         self.queue_draw()
 
     def previous_color(self):
-        return ch.last_color
+        return self.app.ch.last_color
 
-    def set_color(self,color,exclude=None):
+    def set_color(self, hsv, exclude=None):
         for w in self.widgets:
             if w is not exclude:
-                w.set_color(color)
-        self.color = color
-        self.on_select(color)
+                w.set_color(hsv)
+        self.color = hsv_to_rgb(*hsv)
+        self.hsv = hsv
+        self.on_select(hsv)
 
-    def drag_data(self, widget, context, x,y, selection, targetType, time):
-        r,g,b,a = struct.unpack('HHHH', selection.data)
-        clr = (r/65536.0, g/65536.0, b/65536.0)
-        self.set_color(clr)
+    def rgb_selected(self, hsv):
+        self.set_color(hsv, exclude=self.rgb_selector)
 
-    def rgb_selected(self, color):
-        self.set_color(color, exclude=self.rgb_selector)
+    def hsv_selected(self, hsv):
+        self.set_color(hsv, exclude=self.hsv_selector)
 
-    def hsv_selected(self, color):
-        self.set_color(color, exclude=self.hsv_selector)
+    def hue_selected(self, hsv):
+        self.set_color(hsv, exclude=self.circle)
 
-    def hue_selected(self, color):
-        self.set_color(color, exclude=self.circle)
+    def recent_selected(self, hsv):
+        self.set_color(hsv, exclude=self.recent)
 
-    def recent_selected(self, color):
-        self.set_color(color, exclude=self.recent)
-
-    def on_select(self,color):
+    def on_select(self, hsv):
         pass
 
     def expander_expanded_cb(self, expander, junk, cfg_stem):
-        if not self.window_mapped:
+        # Save the expander state
+        if not self.expander_prefs_loaded:
             return
-        # After expander activation (both expand and un-expand), don't let the
-        # colour circle shrink smaller than its current size, and force the WM
-        # to shrink-fit the window to the new size.
-        circle_alloc = self.circle.get_allocation()
-        self.circle.set_size_request(circle_alloc.width, circle_alloc.height)
-        self.window_.set_resizable(False)
         self.app.preferences['colorsampler.%s_expanded' % cfg_stem] \
           = bool(expander.get_expanded())
 
-    def window_configured_cb(self, window, event, *param):
-        if not self.window_mapped:
-            return
-        # Undo the expander trick if it's still in effect: the user should be
-        # able to resize the dialog.
-        if not self.window_.get_resizable():
-            self.window_.set_resizable(True)
-            self.circle.set_size_request(*self.CIRCLE_MIN_SIZE)
-
-    def window_mapped_cb(self, window, event, *params):
-        # On the first map event, restore expander state from prefs.
-        if self.window_mapped:
-            return
+    def show_cb(self, widget):
+        # Restore expander state from prefs.
+        # Use "show", not "map", so that it fires before we have a window
+        assert not self.window
+        assert not self.expander_prefs_loaded
+        # If we wait until then, sidebar positions will drift as a result.
         if self.app.preferences.get("colorsampler.history_expanded", False):
             self.exp_history.set_expanded(True)
         if self.app.preferences.get("colorsampler.details_expanded", False):
             self.exp_details.set_expanded(True)
         if self.app.preferences.get("colorsampler.harmonies_expanded", False):
             self.exp_config.set_expanded(True)
-        self.window_mapped = True
+        self.expander_prefs_loaded = True
+
+
+class ToolWidget (Selector):
+
+    stock_id = stock.TOOL_COLOR_SAMPLER
+
+    def __init__(self, app):
+        Selector.__init__(self, app)
+
+        self.app.brush.observers.append(self.brush_modified_cb)
+        self.stop_callback = False
+
+        # The first callback notification happens before the window is initialized
+        self.set_color(app.brush.get_color_hsv())
+
+    def brush_modified_cb(self, settings):
+        if not settings.intersection(('color_h', 'color_s', 'color_v')):
+            return
+        if self.stop_callback:
+            return
+        self.stop_callback = True
+        hsv = self.app.brush.get_color_hsv()
+        self.set_color(hsv)
+        self.stop_callback = False
+
+    def on_select(self, hsv):
+        if self.stop_callback:
+            return
+        self.app.brush.set_color_hsv(hsv)
+
 
 class Window(windowing.SubWindow):
     def __init__(self,app):
         windowing.SubWindow.__init__(self, app)
         self.set_title(_('MyPaint color selector'))
         self.set_role('Color selector')
-        self.connect('delete-event', self.app.hide_window_cb)
         self.selector = Selector(app, self)
         self.selector.on_select = self.on_select
         self.exp_history = self.selector.exp_history
@@ -1037,18 +1221,23 @@ class Window(windowing.SubWindow):
         self.exp_config = self.selector.exp_config
 
         self.add(self.selector)
-        self.app.brush.settings_observers.append(self.brush_modified_cb)
+        self.app.brush.observers.append(self.brush_modified_cb)
         self.stop_callback = False
 
         # The first callback notification happens before the window is initialized
-        self.selector.set_color(app.brush.get_color_rgb())
+        self.selector.set_color(app.brush.get_color_hsv())
 
-    def brush_modified_cb(self):
-        self.stop_callback = True
-        self.selector.set_color(self.app.brush.get_color_rgb())
-        self.stop_callback = False
-
-    def on_select(self, color):
+    def brush_modified_cb(self, settings):
+        if not settings.intersection(('color_h', 'color_s', 'color_v')):
+            return
         if self.stop_callback:
             return
-        self.app.brush.set_color_rgb(color)
+        self.stop_callback = True
+        hsv = self.app.brush.get_color_hsv()
+        self.selector.set_color(hsv)
+        self.stop_callback = False
+
+    def on_select(self, hsv):
+        if self.stop_callback:
+            return
+        self.app.brush.set_color_hsv(hsv)
