@@ -14,7 +14,7 @@ last_cursor_info = None
 last_cursor = None
 max_cursor_size = None
 
-def get_brush_cursor(radius, is_eraser):
+def get_brush_cursor(radius, is_eraser, is_lockalpha):
     global last_cursor, last_cursor_info, max_cursor_size
 
     if not max_cursor_size:
@@ -22,11 +22,11 @@ def get_brush_cursor(radius, is_eraser):
         max_cursor_size = max(display.get_maximal_cursor_size())
 
     d = int(radius)*2
-    if d < 6: d = 6
+    if d < 3: d = 3
     if is_eraser and d < 8: d = 8
     if d+1 > max_cursor_size:
         d = max_cursor_size-1
-    cursor_info = (d, is_eraser)
+    cursor_info = (d, is_eraser, is_lockalpha)
     if cursor_info != last_cursor_info:
         last_cursor_info = cursor_info
 
@@ -51,6 +51,9 @@ def get_brush_cursor(radius, is_eraser):
             thickness = d/8
             mask.draw_rectangle(bgc, True, d/2-thickness, 0, 2*thickness+1, d+1)
             mask.draw_rectangle(bgc, True, 0, d/2-thickness, d+1, 2*thickness+1)
+        elif is_lockalpha:
+            thickness = int(d/4+0.5)
+            mask.draw_rectangle(bgc, True, d/2-thickness, 0, 2*thickness+1, d+1)
 
         last_cursor = gdk.Cursor(cursor,mask,gdk.color_parse('black'), gdk.color_parse('white'),(d+1)/2,(d+1)/2)
 
