@@ -18,23 +18,26 @@ from bisect import bisect_left
 import lib.document
 from lib import backgroundsurface, command, helpers, layer
 import tileddrawwidget, stategroup
-
+from brushmanager import ManagedBrush
+import stock
 
 class Document(object):
-    def __init__(self, app):
+
+    def __init__(self, app, leader=None):
         self.app = app
-        self.model = lib.document.Document()
-        self.model.set_brush(self.app.brush)
+        self.model = lib.document.Document(self.app.brush)
+        self.followers = []
 
         # View
-        self.tdw = tileddrawwidget.TiledDrawWidget(self.model)
+        self.tdw = tileddrawwidget.TiledDrawWidget(self.app, self.model)
+        self.model.frame_observers.append(self.frame_changed_cb)
 
         # FIXME: hack, to be removed
         fname = os.path.join(self.app.datapath, 'backgrounds', '03_check1.png')
         pixbuf = gdk.pixbuf_new_from_file(fname)
         self.tdw.neutral_background_pixbuf = backgroundsurface.Background(pixbuf)
 
-        self.zoomlevel_values = [1.0/8, 2.0/11, 0.25, 1.0/3, 0.50, 2.0/3,  # micro
+        self.zoomlevel_values = [1.0/16, 1.0/8, 2.0/11, 0.25, 1.0/3, 0.50, 2.0/3,  # micro
                                  1.0, 1.5, 2.0, 3.0, 4.0, 5.5, 8.0,        # normal
                                  11.0, 16.0, 23.0, 32.0, 45.0, 64.0]       # macro
                                  # keep sorted for bisect
@@ -47,38 +50,43 @@ class Document(object):
         self.tdw.zoom_min = min(self.zoomlevel_values)
         self.tdw.zoom_max = max(self.zoomlevel_values)
 
-        # Brush
-        self.app.brush.settings_observers.append(self.brush_modified_cb)
+        # Device change management & pen-stroke watching
         self.tdw.device_observers.append(self.device_changed_cb)
-
+        self.input_stroke_ended_observers.append(self.input_stroke_ended_cb)
         self.last_pen_device = None
-        self.eraser_mode_radius_change = 3*(0.3) # can go back to exact original with brush_smaller_cb()
-        self.eraser_mode_original_radius = None
 
-        self.init_actions()
-        self.init_context_actions()
-        self.app.ui_manager.insert_action_group(self.action_group, -1)
-        for action in self.action_group.list_actions():
-            self.app.kbm.takeover_action(action)
         self.init_stategroups()
-        self.init_extra_keys()
+        if leader is not None:
+            # This is a side doc (e.g. the scratchpad) which plays follow-the-
+            # leader for some events.
+            assert isinstance(leader, Document)
+            leader.followers.append(self)
+            self.action_group = leader.action_group # hack, but needed by tdw
+        else:
+            # This doc owns the Actions which are (sometimes) passed on to
+            # followers to perform. It's also the main document being worked on
+            # by the user.
+            self.init_actions()
+            self.init_context_actions()
+            for action in self.action_group.list_actions():
+                self.app.kbm.takeover_action(action)
+            self.init_extra_keys()
 
     def init_actions(self):
         # name, stock id, label, accelerator, tooltip, callback
         actions = [
-            ('Undo',               gtk.STOCK_UNDO, _('Undo'), 'Z', None, self.undo_cb),
-            ('Redo',               gtk.STOCK_REDO, _('Redo'), 'Y', None, self.redo_cb),
+            ('Undo', gtk.STOCK_UNDO, None, 'Z', None, self.undo_cb),
+            ('Redo', gtk.STOCK_REDO, None, 'Y', None, self.redo_cb),
 
             ('Brighter',     None, _('Brighter'), None, None, self.brighter_cb),
             ('Smaller',      None, _('Smaller'), 'd', None, self.brush_smaller_cb),
             ('MoreOpaque',   None, _('More Opaque'), 's', None, self.more_opaque_cb),
             ('LessOpaque',   None, _('Less Opaque'), 'a', None, self.less_opaque_cb),
-            ('Eraser',       None, _('Toggle Eraser Mode'), 'e', None, self.eraser_cb), # TODO: make toggle action
             ('PickContext',  None, _('Pick Context (layer, brush and color)'), 'w', None, self.pick_context_cb),
 
             ('Darker',       None, _('Darker'), None, None, self.darker_cb),
-            ('Warmer',       None, _('Warmer'), None, None, self.warmer_cb),
-            ('Cooler',       None, _('Cooler'), None, None, self.cooler_cb),
+            ('IncreaseHue',  None, _('Change Color Hue (counter-clockwise)'), None, None, self.increase_hue_cb),
+            ('DecreaseHue',  None, _('Change Color Hue (clockwise)'), None, None, self.decrease_hue_cb),
             ('Purer',        None, _('Purer'), None, None, self.purer_cb),
             ('Grayer',       None, _('Grayer'), None, None, self.grayer_cb),
             ('Bigger',       None, _('Bigger'), 'f', None, self.brush_bigger_cb),
@@ -102,24 +110,47 @@ class Document(object):
 
             ('ShortcutsMenu', None, _('Shortcuts')),
 
-            ('ResetView',   gtk.STOCK_ZOOM_100, _('Reset (Zoom, Rotation, Mirror)'), 'F12', None, self.reset_view_cb),
-            ('ZoomIn',       gtk.STOCK_ZOOM_IN, _('Zoom In (at cursor)'), 'period', None, self.zoom_cb),
-            ('ZoomOut',      gtk.STOCK_ZOOM_OUT, _('Zoom Out'), 'comma', None, self.zoom_cb),
-            ('RotateLeft',   None, _('Rotate Counterclockwise'), '<control>Left', None, self.rotate_cb),
-            ('RotateRight',  None, _('Rotate Clockwise'), '<control>Right', None, self.rotate_cb),
-            ('MirrorHorizontal', None, _('Mirror Horizontal'), 'i', None, self.mirror_horizontal_cb),
-            ('MirrorVertical', None, _('Mirror Vertical'), 'u', None, self.mirror_vertical_cb),
+            ('ResetView',   gtk.STOCK_ZOOM_FIT, _('Reset and Center'), 'F12',
+                _("Reset Zoom, Rotation and Mirroring, and recenter the document"),
+                self.reset_view_cb),
+            ('ResetMenu',   None, _('Reset')),
+            ('ResetZoom',   gtk.STOCK_ZOOM_100, _('Zoom'), None, None, self.reset_view_cb),
+            ('ResetRotation',   None, _('Rotation'), None, None, self.reset_view_cb),
+            ('ResetMirror', None, _('Mirror'), None, None, self.reset_view_cb),
+            ('ZoomIn',       gtk.STOCK_ZOOM_IN, _('Zoom In'), 'period', 
+                _("Increase magnification"), self.zoom_cb),
+            ('ZoomOut',      gtk.STOCK_ZOOM_OUT, _('Zoom Out'), 'comma',
+                _("Decrease magnification"), self.zoom_cb),
+            ('RotateLeft',   stock.ROTATE_LEFT, None, None,
+                _("Rotate the view left"),
+                self.rotate_cb),
+            ('RotateRight',  stock.ROTATE_RIGHT, None, None,
+                _("Rotate the view right"),
+                self.rotate_cb),
+            ('MirrorHorizontal', stock.MIRROR_HORIZONTAL, None, None,
+                _("Mirror: flip the view left to right"),
+                self.mirror_horizontal_cb),
+            ('MirrorVertical', stock.MIRROR_VERTICAL, None, None,
+                _("Flip the view upside-down"),
+                self.mirror_vertical_cb),
             ('SoloLayer',    None, _('Layer Solo'), 'Home', None, self.solo_layer_cb), # TODO: make toggle action
             ('ToggleAbove',  None, _('Hide Layers Above Current'), 'End', None, self.toggle_layers_above_cb), # TODO: make toggle action
+
+            ('BlendMode',    stock.BRUSH_BLEND_MODES),
         ]
-        ag = self.action_group = gtk.ActionGroup('DocumentActions')
+        self.action_group = gtk.ActionGroup('DocumentActions')
+        ag = self.action_group
+        self.app.add_action_group(ag)
         ag.add_actions(actions)
 
+        self.model.command_stack_observers.append(self.update_command_stack_toolitems)
+        self.update_command_stack_toolitems(self.model.command_stack)
+
         toggle_actions = [
-            # name, stock id, label, accelerator, tooltip, callback, default toggle status
-            ('PrintInputs', None, _('Print Brush Input Values to stdout'), None, None, self.print_inputs_cb),
+            ('PrintInputs', None, _('Print Brush Input Values to Console'), None, None, self.print_inputs_cb),
             ('VisualizeRendering', None, _('Visualize Rendering'), None, None, self.visualize_rendering_cb),
             ('NoDoubleBuffereing', None, _('Disable GTK Double Buffering'), None, None, self.no_double_buffering_cb),
+
             ]
         ag.add_toggle_actions(toggle_actions)
 
@@ -178,6 +209,20 @@ class Document(object):
         kbm.add_extra_key('<control>Left', 'RotateLeft')
         kbm.add_extra_key('<control>Right', 'RotateRight')
 
+    @property
+    def input_stroke_ended_observers(self):
+        """Array of callbacks interested in the end of an input stroke.
+
+        Observers are called with the GTK event as their only argument. This
+        is a good place to listen for "just painted something" events;
+        app.brush will contain everything needed about the input stroke which
+        just ended, in the state in which it ended.
+
+        An input stroke is a single pen-down, draw, pen-up action. This sort of
+        stroke is not the same as a brush engine stroke (see ``lib.document``).
+        """
+        return self.tdw._input_stroke_ended_observers
+
     # GENERIC
     def undo_cb(self, action):
         cmd = self.model.undo()
@@ -198,13 +243,13 @@ class Document(object):
             print "WARNING: empty document, nothing copied"
             return
         else:
-            pixbuf = self.model.layer.surface.render_as_pixbuf(*bbox)
+            pixbuf = self.model.layer.render_as_pixbuf(*bbox)
         cb = gtk.Clipboard()
         cb.set_image(pixbuf)
 
     def paste_cb(self, action):
         cb = gtk.Clipboard()
-        def callback(clipboard, pixbuf, trash):
+        def callback(clipboard, pixbuf, junk):
             if not pixbuf:
                 print 'The clipboard doeas not contain any image to paste!'
                 return
@@ -213,15 +258,21 @@ class Document(object):
             self.model.load_layer_from_pixbuf(pixbuf, x, y)
         cb.request_image(callback)
 
-    def brush_modified_cb(self):
-        # called at every brush setting modification, should return fast
-        self.model.set_brush(self.app.brush)
-
-
     def pick_context_cb(self, action):
+        if not self.tdw.has_pointer:
+            for follower in self.followers:
+                if follower.tdw.has_pointer:
+                    print "passing %s action to %s" % (action.get_name(), follower)
+                    follower.pick_context_cb(action)
+                    return
+            return
         x, y = self.tdw.get_cursor_in_model_coordinates()
         for idx, layer in reversed(list(enumerate(self.model.layers))):
-            alpha = layer.surface.get_alpha (x, y, 5) * layer.effective_opacity
+            if layer.locked:
+                continue
+            if not layer.visible:
+                continue
+            alpha = layer.get_alpha (x, y, 5) * layer.effective_opacity
             if alpha > 0.1:
                 old_layer = self.model.layer
                 self.model.select_layer(idx)
@@ -230,26 +281,24 @@ class Document(object):
 
                 # find the most recent (last) stroke that touches our picking point
                 si = self.model.layer.get_stroke_info_at(x, y)
-
                 if si:
-                    picked_brush = None # TODO: remember and restore...?
-                    self.app.brushmanager.select_brush(picked_brush, si.brush_string)
+                    self.restore_brush_from_stroke_info(si)
                     self.si = si # FIXME: should be a method parameter?
                     self.strokeblink_state.activate(action)
                 return
 
+    def restore_brush_from_stroke_info(self, strokeinfo):
+        mb = ManagedBrush(self.app.brushmanager)
+        mb.brushinfo.load_from_string(strokeinfo.brush_string)
+        self.app.brushmanager.select_brush(mb)
+        self.app.brushmodifier.restore_context_of_selected_brush()
+
     # LAYER
     def clear_layer_cb(self, action):
         self.model.clear_layer()
-        if self.model.is_empty():
-            # the user started a new painting
-            self.app.filehandler.filename = None
 
     def remove_layer_cb(self, action):
         self.model.remove_layer()
-        if self.model.is_empty():
-            # the user started a new painting
-            self.app.filehandler.filename = None
 
     def layer_bg_cb(self, action):
         idx = self.model.layer_idx - 1
@@ -294,7 +343,11 @@ class Document(object):
     def pick_layer_cb(self, action):
         x, y = self.tdw.get_cursor_in_model_coordinates()
         for idx, layer in reversed(list(enumerate(self.model.layers))):
-            alpha = layer.surface.get_alpha (x, y, 5) * layer.effective_opacity
+            if layer.locked:
+                continue
+            if not layer.visible:
+                continue
+            alpha = layer.get_alpha (x, y, 5) * layer.effective_opacity
             if alpha > 0.1:
                 self.model.select_layer(idx)
                 self.layerblink_state.activate(action)
@@ -321,55 +374,37 @@ class Document(object):
         adj.set_value(adj.get_value() / 1.8)
 
     def brighter_cb(self, action):
-        self.end_eraser_mode()
         h, s, v = self.app.brush.get_color_hsv()
         v += 0.08
         if v > 1.0: v = 1.0
         self.app.brush.set_color_hsv((h, s, v))
 
     def darker_cb(self, action):
-        self.end_eraser_mode()
         h, s, v = self.app.brush.get_color_hsv()
         v -= 0.08
         # stop a little higher than 0.0, to avoid resetting hue to 0
         if v < 0.005: v = 0.005
         self.app.brush.set_color_hsv((h, s, v))
 
-    def warmer_cb(self,action):
-        self.end_eraser_mode()
+    def increase_hue_cb(self,action):
         h, s, v = self.app.brush.get_color_hsv()
-        # using about 40 degrees (0.111 +- e) as warmest and
-        # 130 deg (0.611 +- e) as coolest
         e = 0.015
-        if 0.111 + e < h < 0.611:
-            h -= 0.015
-        elif 0.0 <= h < 0.111 - e or 0.611<= h <= 1.0:
-            h += 0.015
-        if h > 1.0: h -= 1.0
-        if h < 0.0: h += 1.0
+        h = (h + e) % 1.0
         self.app.brush.set_color_hsv((h, s, v))
 
-    def cooler_cb(self,action):
-        self.end_eraser_mode()
+    def decrease_hue_cb(self,action):
         h, s, v = self.app.brush.get_color_hsv()
         e = 0.015
-        if 0.111 < h < 0.611 - e:
-            h += 0.015
-        elif 0.0 <= h <= 0.111 or 0.611 + e < h <= 1.0:
-            h -= 0.015
-        if h > 1.0: h -= 1.0
-        if h < 0.0: h += 1.0
+        h = (h - e) % 1.0
         self.app.brush.set_color_hsv((h, s, v))
 
     def purer_cb(self,action):
-        self.end_eraser_mode()
         h, s, v = self.app.brush.get_color_hsv()
         s += 0.08
         if s > 1.0: s = 1.0
         self.app.brush.set_color_hsv((h, s, v))
 
     def grayer_cb(self,action):
-        self.end_eraser_mode()
         h, s, v = self.app.brush.get_color_hsv()
         s -= 0.08
         # stop a little higher than 0.0, to avoid resetting hue to 0
@@ -394,7 +429,7 @@ class Document(object):
             context = bm.contexts[i]
         bm.selected_context = context
         if store:
-            context.settings_str = self.app.brush.save_to_string()
+            context.brushinfo = self.app.brush.clone()
             context.preview = bm.selected_brush.preview
             context.save()
         else:
@@ -404,7 +439,7 @@ class Document(object):
             self.app.brush.set_color_hsv(color)
 
     # TDW view manipulation
-    def dragfunc_translate(self, dx, dy):
+    def dragfunc_translate(self, dx, dy, x, y):
         self.tdw.scroll(-dx, -dy)
         # Accelerated scrolling: Highly comfortable to me, but lots of
         # negative feedback from users.  Certainly should be disabled
@@ -413,24 +448,42 @@ class Document(object):
         #
         #self.tdw.scroll(-dx*3, -dy*3)
 
-    def dragfunc_rotate(self, dx, dy):
-        self.tdw.rotate(2*math.pi*dx/300.0)
+    def dragfunc_rotate(self, dx, dy, x, y):
+        # calculate angular velocity from viewport center
+        cx, cy = self.tdw.get_center()
+        x, y = x-cx, y-cy
+        phi2 = math.atan2(y, x)
+        x, y = x-dx, y-dy
+        phi1 = math.atan2(y, x)
+        self.tdw.rotate(phi2-phi1)
+        #self.tdw.rotate(2*math.pi*dx/300.0)
 
-    def dragfunc_zoom(self, dx, dy):
+    def dragfunc_zoom(self, dx, dy, x, y):
         # workaround (should zoom at x=(first click point).x instead of cursor)
         self.tdw.scroll(-dx, -dy)
-        # The meaning of drag direction up/down is a convention from
-        # Blender and probably other 3D tools (Google Earth at least).
-        self.tdw.zoom(math.exp(-dy/100.0))
+        self.tdw.zoom(math.exp(dy/100.0))
 
-    #def dragfunc_rotozoom(self, dx, dy):
+    #def dragfunc_rotozoom(self, dx, dy, x, y):
     #    self.tdw.scroll(-dx, -dy)
     #    self.tdw.zoom(math.exp(-dy/100.0))
     #    self.tdw.rotate(2*math.pi*dx/500.0)
 
+    def dragfunc_frame(self, dx, dy, x, y):
+        if not self.model.frame_enabled:
+            return
+
+        x, y, w, h = self.model.get_frame()
+
+        # Find the difference in document coordinates
+        cr = self.tdw.get_model_coordinates_cairo_context()
+        x0, y0 = cr.device_to_user(x, y)
+        x1, y1 = cr.device_to_user(x+dx, y+dy)
+
+        self.model.move_frame(dx=x1-x0, dy=y1-y0)
+
     def strokeblink_state_enter(self):
         l = layer.Layer()
-        self.si.render_overlay(l.surface)
+        self.si.render_overlay(l)
         self.tdw.overlay_layer = l
         self.tdw.queue_draw() # OPTIMIZE: excess
     def strokeblink_state_leave(self, reason):
@@ -496,12 +549,22 @@ class Document(object):
         self.tdw.mirror()
 
     def reset_view_cb(self, command):
-        self.tdw.set_rotation(0.0)
-        default_zoom = self.app.preferences['view.default_zoom']
-        self.zoomlevel = self.zoomlevel_values.index(default_zoom)
-        self.tdw.set_zoom(default_zoom)
-        self.tdw.set_mirrored(False)
-        self.tdw.recenter_document()
+        if command is None:
+            command_name = None
+            reset_all = True
+        else:
+            command_name = command.get_name()
+            reset_all = (command_name is None) or ('View' in command_name)
+        if reset_all or ('Rotation' in command_name):
+            self.tdw.set_rotation(0.0)
+        if reset_all or ('Zoom' in command_name):
+            default_zoom = self.app.preferences['view.default_zoom']
+            self.zoomlevel = self.zoomlevel_values.index(default_zoom)
+            self.tdw.set_zoom(default_zoom)
+        if reset_all or ('Mirror' in command_name):
+            self.tdw.set_mirrored(False)
+        if reset_all:
+            self.tdw.recenter_document()
 
     # DEBUGGING
     def print_inputs_cb(self, action):
@@ -511,40 +574,17 @@ class Document(object):
     def no_double_buffering_cb(self, action):
         self.tdw.set_double_buffered(not action.get_active())
 
+    # BLEND MODES
+    def clone_selected_brush_for_saving(self):
+        # Clones the current brush, along with blend mode settings.
+        return self.app.brushmanager.clone_selected_brush(name=None)
 
-    # ERASER
-    def eraser_cb(self, action):
-        adj = self.app.brush_adjustment['eraser']
-        if adj.get_value() > 0.9:
-            self.end_eraser_mode()
-        else:
-            # enter eraser mode
-            adj.set_value(1.0)
-            adj2 = self.app.brush_adjustment['radius_logarithmic']
-            r = adj2.get_value()
-            self.eraser_mode_original_radius = r
-            adj2.set_value(r + self.eraser_mode_radius_change)
-
-    def end_eraser_mode(self):
-        adj = self.app.brush_adjustment['eraser']
-        if not adj.get_value() > 0.9:
-            return
-        adj.set_value(0.0)
-        if self.eraser_mode_original_radius:
-            # save eraser radius, restore old radius
-            adj2 = self.app.brush_adjustment['radius_logarithmic']
-            r = adj2.get_value()
-            self.eraser_mode_radius_change = r - self.eraser_mode_original_radius
-            adj2.set_value(self.eraser_mode_original_radius)
-            self.eraser_mode_original_radius = None
+    def device_is_eraser(self, device):
+        if device is None: return False
+        return device.source == gdk.SOURCE_ERASER or 'eraser' in device.name.lower()
 
     def device_changed_cb(self, old_device, new_device):
-        # just enable eraser mode for now (TODO: remember full tool settings)
         # small problem with this code: it doesn't work well with brushes that have (eraser not in [1.0, 0.0])
-        def is_eraser(device):
-            if device is None: return False
-            return device.source == gdk.SOURCE_ERASER or 'eraser' in device.name.lower()
-
         print 'device change:', new_device.name, new_device.source
 
         # When editing brush settings, it is often more convenient to use the mouse.
@@ -559,20 +599,62 @@ class Document(object):
 
         bm = self.app.brushmanager
         if old_device:
-            bm.brush_by_device[old_device.name] = (bm.selected_brush, self.app.brush.save_to_string())
+            old_brush = self.clone_selected_brush_for_saving()
+            bm.store_brush_for_device(old_device.name, old_brush)
 
-
-        if new_device.name in bm.brush_by_device:
-            brush_to_select, brush_settings = bm.brush_by_device[new_device.name]
-            # mark as selected in brushlist, and restore modifications (radius
-            # and/or color changes the user made)
-            bm.select_brush(brush_to_select, brush_settings)
+        if new_device.source == gdk.SOURCE_MOUSE:
+            # Avoid fouling up unrelated devbrushes at stroke end
+            self.app.preferences.pop('devbrush.last_used', None)
         else:
-            # first time using that device
-            adj = self.app.brush_adjustment['eraser']
-            if is_eraser(new_device):
-                # enter eraser mode
-                adj.set_value(1.0)
-            elif not is_eraser(new_device) and is_eraser(old_device):
-                # leave eraser mode
-                adj.set_value(0.0)
+            # Select the brush and update the UI.
+            # Use a sane default if there's nothing associated
+            # with the device yet.
+            brush = bm.fetch_brush_for_device(new_device.name)
+            if brush is None:
+                if self.device_is_eraser(new_device):
+                    brush = bm.get_default_eraser()
+                else:
+                    brush = bm.get_default_brush()
+            self.app.preferences['devbrush.last_used'] = new_device.name
+            bm.select_brush(brush)
+
+    def input_stroke_ended_cb(self, event):
+        # Store device-specific brush settings at the end of the stroke, not
+        # when the device changes because the user can change brush radii etc.
+        # in the middle of a stroke, and because device_changed_cb won't
+        # respond when the user fiddles with colours, opacity and sizes via the
+        # dialogs.
+        device_name = self.app.preferences.get('devbrush.last_used', None)
+        if device_name is None:
+            return
+        selected_brush = self.clone_selected_brush_for_saving()
+        self.app.brushmanager.store_brush_for_device(device_name, selected_brush)
+        # However it may be better to reflect any brush settings change into
+        # the last-used devbrush immediately. The UI idea here is that the
+        # pointer (when you're holding the pen) is special, it's the point of a
+        # real-world tool that you're dipping into a palette, or modifying
+        # using the sliders.
+
+    def update_command_stack_toolitems(self, stack):
+        ag = self.action_group
+        undo_action = ag.get_action("Undo")
+        undo_action.set_sensitive(len(stack.undo_stack) > 0)
+        if len(stack.undo_stack) > 0:
+            cmd = stack.undo_stack[-1]
+            desc = _("Undo %s") % cmd.display_name
+        else:
+            desc = _("Undo: nothing to undo")
+        undo_action.set_label(desc)
+        undo_action.set_tooltip(desc)
+        redo_action = ag.get_action("Redo")
+        redo_action.set_sensitive(len(stack.redo_stack) > 0)
+        if len(stack.redo_stack) > 0:
+            cmd = stack.redo_stack[-1]
+            desc = _("Redo %s") % cmd.display_name
+        else:
+            desc = _("Redo: nothing to redo")
+        redo_action.set_label(desc)
+        redo_action.set_tooltip(desc)
+
+    def frame_changed_cb(self):
+        self.tdw.queue_draw()
