@@ -10,10 +10,22 @@ import gtk, gobject, cairo, random
 gdk = gtk.gdk
 from math import floor, ceil, pi, log
 from numpy import isfinite
+from warnings import warn
 
 from lib import helpers, tiledsurface, pixbufsurface
 import cursor
 
+
+
+def _make_testbed_model():
+    warn("Creating standalone model for testing", RuntimeWarning, 2)
+    import lib.brush, lib.document
+    brush = lib.brush.BrushInfo()
+    brush.load_defaults()
+    return lib.document.Document(brush)
+
+
+
 class TiledDrawWidget(gtk.DrawingArea):
     """
     This widget displays a document (../lib/document*.py).
@@ -23,15 +35,21 @@ class TiledDrawWidget(gtk.DrawingArea):
     passed to the document after applying the inverse transformation.
     """
 
-    def __init__(self, document):
+    # Register a GType name for Glade, GtkBuilder etc.
+    __gtype_name__ = "TiledDrawWidget"
+
+    CANNOT_DRAW_CURSOR = gdk.Cursor(gdk.CIRCLE)
+
+    def __init__(self, app=None, document=None):
         gtk.DrawingArea.__init__(self)
         self.connect("expose-event", self.expose_cb)
         self.connect("enter-notify-event", self.enter_notify_cb)
         self.connect("leave-notify-event", self.leave_notify_cb)
         self.connect("size-allocate", self.size_allocate_cb)
+        self.connect("state-changed", self.state_changed_cb)
 
         # workaround for https://gna.org/bugs/?14372 ([Windows] crash when moving the pen during startup)
-        def at_application_start(*trash):
+        def at_application_start(*junk):
             self.connect("motion-notify-event", self.motion_notify_cb)
             self.connect("button-press-event", self.button_press_cb)
             self.connect("button-release-event", self.button_release_cb)
@@ -54,9 +72,12 @@ class TiledDrawWidget(gtk.DrawingArea):
 
         self.set_extension_events (gdk.EXTENSION_EVENTS_ALL)
 
+        self.app = app
+        if document is None:
+            document = _make_testbed_model()
         self.doc = document
         self.doc.canvas_observers.append(self.canvas_modified_cb)
-        self.doc.brush.settings_observers.append(self.brush_modified_cb)
+        self.doc.brush.brushinfo.observers.append(self.brush_modified_cb)
 
         self.cursor_info = None
 
@@ -66,7 +87,8 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.last_event_device = None
         self.last_event_had_pressure_info = False
         self.last_painting_pos = None
-        self.device_observers = []
+        self.device_observers = [] #: Notified during drawing when input devices change
+        self._input_stroke_ended_observers = [] #: Access via gui.document
 
         self.visualize_rendering = False
 
@@ -93,13 +115,28 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.bad_devices = []
         self.motions = []
 
-        self.set_sensitive(True)
+        # Sensitivity; we draw via a cached snapshot while the widget is
+        # insensitive. tdws are generally only insensitive during loading and
+        # saving, and because we now process the GTK main loop during loading
+        # and saving, we need to avoid drawing partially-loaded files.
+
+        self.is_sensitive = True    # just mirrors gtk.STATE_INSENSITIVE
+        self.snapshot_pixmap = None
+
+        self.override_cursor = None
 
     #def set_scroll_at_edges(self, choice):
     #    self.scroll_at_edges = choice
 
-    def set_sensitive(self, sensitive):
-        """Set if the widget accepts input or not"""
+    def state_changed_cb(self, widget, oldstate):
+        # Keeps track of the sensitivity state, and regenerates
+        # the snapshot pixbuf on entering it.
+        sensitive = self.get_state() != gtk.STATE_INSENSITIVE
+        if sensitive:
+            self.snapshot_pixmap = None
+        else:
+            if self.snapshot_pixmap is None:
+                self.snapshot_pixmap = self.get_snapshot()
         self.is_sensitive = sensitive
 
     def enter_notify_cb(self, widget, event):
@@ -108,44 +145,55 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.has_pointer = False
 
     def size_allocate_cb(self, widget, allocation):
-        new_size = tuple(allocation)[2:4]
-        old_size = getattr(self, 'current_size', new_size)
-        self.current_size = new_size
-        if new_size != old_size:
-            # recenter
-            dx = old_size[0] - new_size[0]
-            dy = old_size[1] - new_size[1]
-            self.scroll(dx/2, dy/2)
+        old_alloc = getattr(self, 'stored_allocation', allocation)
+        if old_alloc != allocation:
+            dx = allocation.x - old_alloc.x
+            dy = allocation.y - old_alloc.y
+            self.scroll(dx, dy)
+        self.stored_allocation = allocation
 
     def device_used(self, device):
         """Tell the TDW about a device being used."""
         if device == self.last_event_device:
-            return
+            return True
         for func in self.device_observers:
             func(self.last_event_device, device)
         self.last_event_device = device
+        return False
+
+        # Do not interpolate between motion events from different
+        # devices.  If the final pressure value from the previous
+        # device was not 0.0, the motion event of the new device could
+        # cause a visible stroke, even if pressure is 0.0.
+        self.doc.brush.reset()
 
     def motion_notify_cb(self, widget, event, button1_pressed=None):
         if not self.is_sensitive:
             return
-
+        
         if self.last_event_time:
             dtime = (event.time - self.last_event_time)/1000.0
             dx = event.x - self.last_event_x
             dy = event.y - self.last_event_y
         else:
             dtime = None
-        self.device_used(event.device)
         self.last_event_x = event.x
         self.last_event_y = event.y
         self.last_event_time = event.time
         if dtime is None:
             return
+        
+        same_device = self.device_used(event.device)
 
         if self.dragfunc:
-            self.dragfunc(dx, dy)
+            self.dragfunc(dx, dy, event.x, event.y)
             return
 
+        # Refuse drawing if the layer is locked or hidden
+        if self.doc.layer.locked or not self.doc.layer.visible:
+            return
+            # TODO: some feedback, maybe
+
         cr = self.get_model_coordinates_cairo_context()
         x, y = cr.device_to_user(event.x, event.y)
         
@@ -213,6 +261,14 @@ class TiledDrawWidget(gtk.DrawingArea):
         if pressure:
             self.last_painting_pos = x, y
 
+        # If the device has changed and the last pressure value from the previous device
+        # is not equal to 0.0, this can leave a visible stroke on the layer even if the 'new'
+        # device is not pressed on the tablet and has a pressure axis == 0.0.
+        # Reseting the brush when the device changes fixes this issue, but there may be a
+        # much more elegant solution that only resets the brush on this edge-case.
+        if not same_device:
+            self.doc.brush.reset()
+
         # On Windows, GTK timestamps have a resolution around
         # 15ms, but tablet events arrive every 8ms.
         # https://gna.org/bugs/index.php?16569
@@ -245,11 +301,6 @@ class TiledDrawWidget(gtk.DrawingArea):
             return
 
         if event.button == 1:
-            # straight line
-            if (event.state & gdk.SHIFT_MASK) and self.last_painting_pos:
-                dst = self.get_cursor_in_model_coordinates()
-                self.doc.straight_line(self.last_painting_pos, dst)
-
             # mouse button pressed (while painting without pressure information)
             if not self.last_event_had_pressure_info:
                 # For the mouse we don't get a motion event for "pressure"
@@ -261,6 +312,17 @@ class TiledDrawWidget(gtk.DrawingArea):
         # (see comment above in button_press_cb)
         if event.button == 1 and not self.last_event_had_pressure_info:
             self.motion_notify_cb(win, event, button1_pressed=False)
+        # Outsiders can access this via gui.document
+        for func in self._input_stroke_ended_observers:
+            func(event)
+
+    def straight_line_from_last_pos(self, is_sequence=False):
+        if not self.last_painting_pos:
+            return
+        dst = self.get_cursor_in_model_coordinates()
+        self.doc.straight_line(self.last_painting_pos, dst)
+        if is_sequence:
+            self.last_painting_pos = dst
 
     def canvas_modified_cb(self, x1, y1, w, h):
         if not self.window:
@@ -287,8 +349,13 @@ class TiledDrawWidget(gtk.DrawingArea):
 
     def expose_cb(self, widget, event):
         self.update_cursor() # hack to get the initial cursor right
-        #print 'expose', tuple(event.area)
-        self.repaint(event.area)
+        if self.snapshot_pixmap:
+            gc = self.get_style().fg_gc[self.get_state()]
+            area = event.area
+            x,y,w,h = area.x, area.y, area.width, area.height
+            self.window.draw_drawable(gc, self.snapshot_pixmap, x,y, x,y, w,h)
+        else:
+            self.repaint(event.area)
         return True
 
     def get_model_coordinates_cairo_context(self, cr=None):
@@ -339,6 +406,10 @@ class TiledDrawWidget(gtk.DrawingArea):
         return layers
 
     def repaint(self, device_bbox=None):
+        cr, surface, sparse, mipmap_level, gdk_clip_region = self.render_prepare(device_bbox)
+        self.render_execute(cr, surface, sparse, mipmap_level, gdk_clip_region)
+
+    def render_prepare(self, device_bbox):
         if device_bbox is None:
             w, h = self.window.get_size()
             device_bbox = (0, 0, w, h)
@@ -365,8 +436,15 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.get_model_coordinates_cairo_context(cr)
 
         # choose best mipmap
-        mipmap_level = max(0, int(ceil(log(1/self.scale,2))))
-        #mipmap_level = max(0, int(floor(log(1.0/self.scale,2)))) # slightly better quality but clearly slower
+        hq_zoom = False
+        if self.app and self.app.preferences['view.high_quality_zoom']:
+            hq_zoom = True
+        if hq_zoom:
+            # can cause a very clear slowdown on some hardware
+            # (we probably could avoid this by doing rendering differently)
+            mipmap_level = max(0, int(floor(log(1.0/self.scale,2))))
+        else:
+            mipmap_level = max(0, int(ceil(log(1/self.scale,2))))
         # OPTIMIZE: if we would render tile scanlines, we could probably use the better one above...
         mipmap_level = min(mipmap_level, tiledsurface.MAX_MIPMAP_LEVEL)
         cr.scale(2**mipmap_level, 2**mipmap_level)
@@ -393,7 +471,12 @@ class TiledDrawWidget(gtk.DrawingArea):
 
         del x1, y1, x2, y2, w, h
 
+        return cr, surface, sparse, mipmap_level, gdk_clip_region
+
+    def render_execute(self, cr, surface, sparse, mipmap_level, gdk_clip_region):
+        translation_only = self.is_translation_only()
         model_bbox = surface.x, surface.y, surface.w, surface.h
+
         #print 'model bbox', model_bbox
 
         # not sure if it is a good idea to clip so tightly
@@ -463,6 +546,21 @@ class TiledDrawWidget(gtk.DrawingArea):
 
             cr.paint()
 
+        if self.doc.frame_enabled:
+            # Draw a semi-transparent black overlay for
+            # all the area outside the "document area"
+            cr.save()
+            cr.set_source_rgba(0, 0, 0, 0.6)
+            cr.set_operator(cairo.OPERATOR_OVER)
+            mipmap_factor = 2**mipmap_level
+            frame = self.doc.get_frame()
+            cr.rectangle(frame[0]/mipmap_factor, frame[1]/mipmap_factor,
+                            frame[2]/mipmap_factor, frame[3]/mipmap_factor)
+            cr.rectangle(*model_bbox)
+            cr.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
+            cr.fill()
+            cr.restore()
+
         if self.visualize_rendering:
             # visualize painted bboxes (blue)
             cr.set_source_rgba(0, 0, random.random(), 0.4)
@@ -484,12 +582,16 @@ class TiledDrawWidget(gtk.DrawingArea):
         else:
             self.queue_draw()
 
+    def get_center(self):
+        w, h = self.window.get_size()
+        return w/2.0, h/2.0
+
     def rotozoom_with_center(self, function, at_pointer=False):
         if at_pointer and self.has_pointer and self.last_event_x is not None:
             cx, cy = self.last_event_x, self.last_event_y
         else:
             w, h = self.window.get_size()
-            cx, cy = w/2.0, h/2.0
+            cx, cy = self.get_center()
         cr = self.get_model_coordinates_cairo_context()
         cx_device, cy_device = cr.device_to_user(cx, cy)
         function()
@@ -534,7 +636,7 @@ class TiledDrawWidget(gtk.DrawingArea):
             self.dragfunc = None
 
     def recenter_document(self):
-        x, y, w, h = self.doc.get_bbox()
+        x, y, w, h = self.doc.get_effective_bbox()
         desired_cx_user = x+w/2
         desired_cy_user = y+h/2
 
@@ -546,18 +648,45 @@ class TiledDrawWidget(gtk.DrawingArea):
         self.translation_y += (cy_user - desired_cy_user)*self.scale
         self.queue_draw()
 
-    def brush_modified_cb(self):
+    def brush_modified_cb(self, settings):
         self.update_cursor()
 
     def update_cursor(self):
-        if not self.window: return
-
-        b = self.doc.brush
-        radius = b.get_actual_radius()*self.scale
-        c = cursor.get_brush_cursor(radius, b.is_eraser())
+        if not self.window:
+            return
+        elif self.override_cursor is not None:
+            c = self.override_cursor
+        elif not self.is_sensitive:
+            c = None
+        elif self.doc.layer.locked or not self.doc.layer.visible:
+            c = self.CANNOT_DRAW_CURSOR
+        else:
+            b = self.doc.brush.brushinfo
+            radius = b.get_effective_radius()*self.scale
+            c = cursor.get_brush_cursor(radius, b.is_eraser(), b.get_base_value('lock_alpha') > 0.9)
         self.window.set_cursor(c)
 
+    def set_override_cursor(self, cursor):
+        """Set a cursor which will always be used.
+
+        Used by the colour picker. The override cursor will be used regardless
+        of the criteria update_cursor() normally uses. Pass None to let it
+        choose normally again.
+        """
+        self.override_cursor = cursor
+        self.update_cursor()
+
     def toggle_show_layers_above(self):
         self.show_layers_above = not self.show_layers_above
         self.queue_draw()
 
+
+if __name__ == '__main__':
+    tdw = TiledDrawWidget()
+    tdw.set_size_request(640, 480)
+    win = gtk.Window()
+    win.set_title("tdw test")
+    win.connect("destroy", lambda *a: gtk.main_quit())
+    win.add(tdw)
+    win.show_all()
+    gtk.main()
