@@ -1,9 +1,12 @@
 #include "Python.h"
 #include "numpy/arrayobject.h"
 #include "../brushlib/brushlib.hpp"
+#include "python_brush.hpp"
+#include "brushmodes.hpp"
 #include "tiledsurface.hpp"
 #include "pixops.hpp"
 #include "colorring.hpp"
-#include "colorchanger.hpp"
+#include "colorchanger_wash.hpp"
+#include "colorchanger_crossed_bowl.hpp"
 #include "gdkpixbuf2numpy.hpp"
 #include "fastpng.hpp"
